import os
import pickle
import time
import urllib.request

import pandas as pd
import pytest
from perpetual import PerpetualBooster

RESOURCES_DIR = os.path.join(os.path.dirname(__file__), "..", "..", "..", "resources")

DOWNLOAD_URLS = {
    "tabarena_x.csv": "https://github.com/user-attachments/files/25580857/X.csv",
    "tabarena_y.csv": "https://github.com/user-attachments/files/25580858/y.csv",
    "tabarena_cat_x.csv": "https://github.com/user-attachments/files/25600227/X.csv",
    "tabarena_cat_y.csv": "https://github.com/user-attachments/files/25600228/y.csv",
    "tabarena_wide_x.csv": "https://github.com/user-attachments/files/25682565/X.csv",
    "tabarena_wide_y.csv": "https://github.com/user-attachments/files/25682566/y.csv",
    "tabarena_var_x.csv": "https://github.com/user-attachments/files/25739177/X.csv",
    "tabarena_var_y.csv": "https://github.com/user-attachments/files/25739178/y.csv",
    "tabarena_load_x.csv": "https://github.com/user-attachments/files/25764663/X.csv",
    "tabarena_load_y.csv": "https://github.com/user-attachments/files/25764665/y.csv",
}


def _get_csv(filename):
    """Try to read a CSV from resources dir, then download, or skip the test."""
    print(f"Loading {filename}...")
    local_path = os.path.join(RESOURCES_DIR, filename)
    if os.path.isfile(local_path):
        res = pd.read_csv(local_path)
        print(f"Loaded {filename} from resources.")
        return res

    # Try downloading to the resources folder
    url = DOWNLOAD_URLS.get(filename)
    if url is None:
        pytest.skip(f"No download URL configured for {filename}")
    try:
        print(f"Downloading {filename} from {url}...")
        os.makedirs(RESOURCES_DIR, exist_ok=True)
        urllib.request.urlretrieve(url, local_path)
        print(f"Downloaded {filename}.")
    except Exception as exc:
        pytest.skip(f"Could not download {filename}: {exc}")

    return pd.read_csv(local_path)


def test_tabarena():
    X = _get_csv("tabarena_x.csv")
    y = _get_csv("tabarena_y.csv")

    model = PerpetualBooster(budget=1.0, categorical_features=["state", "area_code"])

    model.fit(X, y)


def test_tabarena_categorical():
    X = _get_csv("tabarena_cat_x.csv")
    y = _get_csv("tabarena_cat_y.csv")

    categorical_features = [
        "RESOURCE",
        "MGR_ID",
        "ROLE_ROLLUP_1",
        "ROLE_ROLLUP_2",
        "ROLE_DEPTNAME",
        "ROLE_TITLE",
        "ROLE_FAMILY_DESC",
        "ROLE_FAMILY",
    ]

    model = PerpetualBooster(categorical_features=categorical_features)
    model.fit(X, y)

    model = PerpetualBooster(categorical_features=categorical_features, max_cat=3000)
    model.fit(X, y)


def test_tabarena_wide():
    X = _get_csv("tabarena_wide_x.csv")
    y = _get_csv("tabarena_wide_y.csv")

    categorical_features = [
        "INF_ANAM",
        "STENOK_AN",
        "FK_STENOK",
        "IBS_POST",
        "IBS_NASL",
        "GB",
        "SIM_GIPERT",
        "DLIT_AG",
        "ZSN_A",
        "nr_11",
        "nr_01",
        "nr_02",
        "nr_03",
        "nr_04",
        "nr_07",
        "nr_08",
        "np_01",
        "np_04",
        "np_05",
        "np_08",
        "np_09",
        "endocr_01",
        "endocr_02",
        "endocr_03",
        "zab_leg_01",
        "zab_leg_02",
        "zab_leg_03",
        "zab_leg_04",
        "zab_leg_06",
        "O_L_POST",
        "K_SH_POST",
        "MP_TP_POST",
        "SVT_POST",
        "GT_POST",
        "FIB_G_POST",
        "ant_im",
        "lat_im",
        "inf_im",
        "post_im",
        "IM_PG_P",
        "ritm_ecg_p_01",
        "ritm_ecg_p_02",
        "ritm_ecg_p_04",
        "ritm_ecg_p_07",
        "ritm_ecg_p_08",
        "n_r_ecg_p_01",
        "n_r_ecg_p_02",
        "n_r_ecg_p_03",
        "n_r_ecg_p_04",
        "n_r_ecg_p_05",
        "n_r_ecg_p_06",
        "n_r_ecg_p_08",
        "n_r_ecg_p_09",
        "n_r_ecg_p_10",
        "n_p_ecg_p_01",
        "n_p_ecg_p_03",
        "n_p_ecg_p_04",
        "n_p_ecg_p_05",
        "n_p_ecg_p_06",
        "n_p_ecg_p_07",
        "n_p_ecg_p_08",
        "n_p_ecg_p_09",
        "n_p_ecg_p_10",
        "n_p_ecg_p_11",
        "n_p_ecg_p_12",
        "fibr_ter_01",
        "fibr_ter_02",
        "fibr_ter_03",
        "fibr_ter_05",
        "fibr_ter_06",
        "fibr_ter_07",
        "fibr_ter_08",
        "GIPO_K",
        "GIPER_NA",
        "TIME_B_S",
        "R_AB_1_n",
        "R_AB_2_n",
        "R_AB_3_n",
        "NA_KB",
        "NOT_NA_KB",
        "LID_KB",
        "NITR_S",
        "NOT_NA_1_n",
        "LID_S_n",
        "B_BLOK_S_n",
        "ANT_CA_S_n",
        "GEPAR_S_n",
        "ASP_S_n",
        "TIKL_S_n",
        "TRENT_S_n",
    ]

    model = PerpetualBooster(
        objective="LogLoss",
        budget=2.0,
        categorical_features=categorical_features,
        memory_limit=1,
        iteration_limit=3,
        timeout=60 * 15,
    )

    start_time = time.time()
    model.fit(X, y)
    end_time = time.time()

    print(f"Trees: {model.number_of_trees}")
    print(f"Fit time: {end_time - start_time}")


def test_tabarena_var():
    X = _get_csv("tabarena_var_x.csv")
    y = _get_csv("tabarena_var_y.csv")

    categorical_features = [
        "Var192",
        "Var193",
        "Var194",
        "Var195",
        "Var196",
        "Var197",
        "Var198",
        "Var199",
        "Var200",
        "Var201",
        "Var202",
        "Var203",
        "Var204",
        "Var205",
        "Var206",
        "Var207",
        "Var208",
        "Var210",
        "Var212",
        "Var216",
        "Var217",
        "Var218",
        "Var219",
        "Var221",
        "Var223",
        "Var225",
        "Var226",
        "Var227",
        "Var228",
        "Var229",
    ]

    model = PerpetualBooster(
        categorical_features=categorical_features, iteration_limit=3, memory_limit=1
    )
    model.fit(X, y)


def test_tabarena_save_load():
    print("test_tabarena_save_load started.")
    X = _get_csv("tabarena_load_x.csv")
    y = _get_csv("tabarena_load_y.csv")

    categorical_features = [
        "position_-30",
        "position_-29",
        "position_-28",
        "position_-27",
        "position_-26",
        "position_-25",
        "position_-24",
        "position_-23",
        "position_-22",
        "position_-21",
        "position_-20",
        "position_-19",
        "position_-18",
        "position_-17",
        "position_-16",
        "position_-15",
        "position_-14",
        "position_-13",
        "position_-12",
        "position_-11",
        "position_-10",
        "position_-9",
        "position_-8",
        "position_-7",
        "position_-6",
        "position_-5",
        "position_-4",
        "position_-3",
        "position_-2",
        "position_-1",
        "position_1",
        "position_2",
        "position_3",
        "position_4",
        "position_5",
        "position_6",
        "position_7",
        "position_8",
        "position_9",
        "position_10",
        "position_11",
        "position_12",
        "position_13",
        "position_14",
        "position_15",
        "position_16",
        "position_17",
        "position_18",
        "position_19",
        "position_20",
        "position_21",
        "position_22",
        "position_23",
        "position_24",
        "position_25",
        "position_26",
        "position_27",
        "position_28",
        "position_29",
        "position_30",
    ]

    model = PerpetualBooster(
        categorical_features=categorical_features,
        memory_limit=3,
        num_threads=8,
        objective="LogLoss",
        iteration_limit=10,
        budget=2.0,
    )

    print(f"Starting fit... memory_limit={model.memory_limit}")
    model.fit(X, y)
    print("Fit completed.")
    print(f"Number of trees: {model.number_of_trees}")

    model_path = os.path.join(RESOURCES_DIR, "model.pkl")

    print(f"Saving model to {model_path}...")
    with open(model_path, "wb") as f:
        pickle.dump(model, f)
    print("Model saved.")

    print("Loading model...")
    with open(model_path, "rb") as f:
        loaded_model = pickle.load(f)
    print("Model loaded.")

    del loaded_model
    try:
        os.remove(model_path)
    except OSError:
        pass



from __future__ import annotations

import numpy as np
import pytest
from perpetual.serialize import (
    NumpySerializer,
    ObjectItem,
    ObjectSerializer,
    Scaler,
    ScalerSerializer,
)

scaler_values = [
    1,
    1.0,
    1.00101,
    "a string",
    True,
    False,
    None,
]


@pytest.mark.parametrize("value", scaler_values)
def test_scaler(value: Scaler):
    serializer = ScalerSerializer()
    r = serializer.serialize(value)
    assert isinstance(r, str)
    assert value == serializer.deserialize(r)


object_values = [
    [1, 2, 3],
    [1.0, 4.0],
    ["a", "b", "c"],
    {"a": 1.0, "b": 2.0},
    {"a": "test", "b": "what"},
    *scaler_values,
]


@pytest.mark.parametrize("value", object_values)
def test_object(value: ObjectItem):
    serializer = ObjectSerializer()
    r = serializer.serialize(value)
    assert isinstance(r, str)
    assert value == serializer.deserialize(r)


numpy_values = [
    np.array([1.0, 2.23]),
    np.array([1, 2, 3, 4, 5, 6]).reshape((2, 3)),
    np.array([1, 2, 3, 4, 5, 6], dtype="int").reshape((2, 3)),
]


@pytest.mark.parametrize("value", numpy_values)
def test_numpy(value: np.ndarray):
    serializer = NumpySerializer()
    r = serializer.serialize(value)
    assert isinstance(r, str)
    assert np.array_equal(value, serializer.deserialize(r))


def test_serialization_module():
    # ScalerSerializer
    ss = ScalerSerializer()
    assert ss.serialize(1) == "1"
    assert ss.serialize("test") == "'test'"
    assert ss.deserialize("1") == 1
    assert ss.deserialize("'test'") == "test"

    # ObjectSerializer
    os_ser = ObjectSerializer()
    obj = {"a": [1, 2], "b": 3.0}
    ser = os_ser.serialize(obj)
    assert os_ser.deserialize(ser) == obj
    # Test numpy array handled in ObjectSerializer
    assert os_ser.serialize(np.array([1, 2])) == "[1, 2]"

    # NumpySerializer
    ns = NumpySerializer()
    a1 = np.array([1, 2, 3], dtype=np.float64)
    ser1 = ns.serialize(a1)
    assert np.allclose(ns.deserialize(ser1), a1)

    a2 = np.array([[1, 2], [3, 4]], dtype=np.int32)
    ser2 = ns.serialize(a2)
    a2_deser = ns.deserialize(ser2)
    assert a2_deser.shape == (2, 2)
    assert np.array_equal(a2_deser, a2)



"""Serialization helpers for persisting model metadata.

Provides abstract and concrete serializers for scalars, JSON objects,
and NumPy arrays.
"""

from __future__ import annotations

import json
from abc import ABC, abstractmethod
from ast import literal_eval
from dataclasses import dataclass
from typing import Dict, Generic, List, Tuple, TypeVar, Union

import numpy as np
import numpy.typing as npt

T = TypeVar("T")


class BaseSerializer(ABC, Generic[T]):
    """Abstract base for metadata serializers."""

    @abstractmethod
    def serialize(self, obj: T) -> str:
        """serialize method - should take an object and return a string"""

    @abstractmethod
    def deserialize(self, obj_repr: str) -> T:
        """deserialize method - should take a string and return original object"""


Scaler = Union[int, float, str]


class ScalerSerializer(BaseSerializer[Scaler]):
    """Serializer for scalar values (int, float, str)."""

    def serialize(self, obj: Scaler) -> str:
        if isinstance(obj, str):
            obj_ = f"'{obj}'"
        else:
            obj_ = str(obj)
        return obj_

    def deserialize(self, obj_repr: str) -> Scaler:
        return literal_eval(node_or_string=obj_repr)


ObjectItem = Union[
    List[Scaler],
    Dict[str, Scaler],
    Scaler,
]


class ObjectSerializer(BaseSerializer[ObjectItem]):
    """Serializer for JSON-compatible objects (lists, dicts, scalars)."""

    def serialize(self, obj: ObjectItem) -> str:
        if isinstance(obj, np.ndarray):
            obj = obj.tolist()
        return json.dumps(obj)

    def deserialize(self, obj_repr: str) -> ObjectItem:
        return json.loads(obj_repr)


@dataclass
class NumpyData:
    """Intermediate representation for a serialized NumPy array."""

    array: Union[List[float], List[int]]
    dtype: str
    shape: Tuple[int, ...]


class NumpySerializer(BaseSerializer[npt.NDArray]):
    """Serializer that round-trips NumPy arrays through JSON."""

    def serialize(self, obj: npt.NDArray) -> str:
        return json.dumps(
            {"array": obj.tolist(), "dtype": str(obj.dtype), "shape": obj.shape}
        )

    def deserialize(self, obj_repr: str) -> npt.NDArray:
        data = NumpyData(**json.loads(obj_repr))
        a = np.array(data.array, dtype=data.dtype)  # type: ignore
        if len(data.shape) == 1:
            return a
        else:
            return a.reshape(data.shape)



import os
import pickle

import numpy as np
import pandas as pd
from perpetual import PerpetualBooster


def run_model_io_example():
    # 1. Prepare some data
    print("Preparing data...")
    from sklearn.datasets import make_regression

    X, y = make_regression(n_samples=1000, n_features=10, noise=0.1, random_state=42)
    feature_names = [f"feature_{i}" for i in range(10)]
    X_df = pd.DataFrame(X, columns=feature_names)

    # 2. Fit the model
    print("Fitting PerpetualBooster...")
    model = PerpetualBooster(objective="SquaredLoss", budget=0.1)
    model.fit(X_df, y)

    # Add some custom metadata
    model.insert_metadata("author", "Perpetual User")
    model.insert_metadata("dataset", "Synthetic Regression")

    original_preds = model.predict(X_df[:5])
    print(f"Original predictions: {original_preds}")

    # 3. Native Save/Load
    print("\nTesting Native Save/Load...")
    model_path = "model.json"
    model.save_booster(model_path)

    loaded_model = PerpetualBooster.load_booster(model_path)
    loaded_preds = loaded_model.predict(X_df[:5])
    print(f"Loaded predictions: {loaded_preds}")
    print(f"Metadata 'author': {loaded_model.get_metadata('author')}")

    assert np.allclose(original_preds, loaded_preds)
    os.remove(model_path)

    # 4. Pickle Save/Load
    print("\nTesting Pickle Save/Load...")
    pickle_path = "model.pkl"
    with open(pickle_path, "wb") as f:
        pickle.dump(model, f)

    with open(pickle_path, "rb") as f:
        pickled_model = pickle.load(f)

    pickled_preds = pickled_model.predict(X_df[:5])
    print(f"Pickled predictions: {pickled_preds}")
    assert np.allclose(original_preds, pickled_preds)
    os.remove(pickle_path)

    # 5. Export to XGBoost JSON
    print("\nExporting to XGBoost format...")
    xgb_path = "model_xgb.json"
    model.save_as_xgboost(xgb_path)
    print(f"XGBoost model saved to {xgb_path}")

    # If xgboost is installed, we could verify it here
    try:
        import xgboost as xgb

        bst = xgb.Booster()
        bst.load_model(xgb_path)
        dmat = xgb.DMatrix(X_df[:5])
        xgb_preds = bst.predict(dmat)
        print(f"XGBoost predictions: {xgb_preds}")
    except ImportError:
        print("XGBoost not installed, skipping verification.")

    if os.path.exists(xgb_path):
        os.remove(xgb_path)

    # 6. Export to ONNX
    print("\nExporting to ONNX format...")
    onnx_path = "model.onnx"
    model.save_as_onnx(onnx_path)
    print(f"ONNX model saved to {onnx_path}")

    # If onnxruntime is installed, we could verify it here
    try:
        import onnxruntime as rt

        sess = rt.InferenceSession(onnx_path)
        input_name = sess.get_inputs()[0].name
        X_f32 = X_df[:5].to_numpy().astype(np.float32)
        onnx_preds = sess.run(None, {input_name: X_f32})[0].flatten()
        print(f"ONNX predictions: {onnx_preds}")
    except ImportError:
        print("onnxruntime not installed, skipping verification.")

    if os.path.exists(onnx_path):
        os.remove(onnx_path)

    print("\nExample completed successfully!")


if __name__ == "__main__":
    run_model_io_example()



import numpy as np
import pytest
from perpetual.dml import DMLEstimator


def test_dml_estimator_native():
    """Test that the native Rust DMLEstimator fits and predicts correctly."""
    # Synthetic data
    np.random.seed(42)
    n = 1000
    X = np.random.normal(0, 1, size=(n, 5))

    # Treatment is continuous or binary, let's use binary with confounding
    # W ~ Bernoulli(sigma(X[:, 0]))
    logits = X[:, 0]
    p = 1 / (1 + np.exp(-logits))
    W = np.random.binomial(1, p)

    # Outcome Y: treatment effect is 2.0 (constant for simplicity natively)
    # y = \theta(X) W + g(X) + \epsilon
    # g(X) = 1.5 * X[:, 1]
    y = 2.0 * W + 1.5 * X[:, 1] + np.random.normal(0, 0.1, size=n)

    # Initialize DMLEstimator
    dml = DMLEstimator(budget=0.5, n_folds=3, seed=42)

    # Fit
    dml.fit(X, W, y)

    # Predict CATE
    cate_pred = dml.predict(X)

    # Check that predictions are close to true effect 2.0
    assert cate_pred.shape == (n,)
    assert np.mean(cate_pred) == pytest.approx(2.0, abs=0.2), (
        f"Mean CATE was {np.mean(cate_pred)}"
    )

    # Test ATE Inference
    inf = dml.ate_inference()

    assert "ate" in inf
    assert "std_err" in inf
    assert "ci_lower" in inf
    assert "ci_upper" in inf

    ate = inf["ate"]
    se = inf["std_err"]
    lower = inf["ci_lower"]
    upper = inf["ci_upper"]

    assert ate == pytest.approx(2.0, abs=0.2)
    assert se > 0
    assert lower < ate < upper

    # Test feature importances
    importances = dml.feature_importances_
    assert importances is not None
    assert len(importances) == 5
    assert isinstance(importances, np.ndarray)



import unittest

import numpy as np
import pandas as pd
from perpetual.uplift import UpliftBooster


class TestUpliftBooster(unittest.TestCase):
    def setUp(self):
        # Generate small synthetic data for testing
        np.random.seed(42)
        self.X = np.random.randn(100, 5)
        self.w = np.random.randint(0, 2, 100)
        self.y = (
            0.5 * self.X[:, 0]
            + self.w * (0.2 * self.X[:, 1])
            + np.random.randn(100) * 0.1
        )

    def test_initialization(self):
        model = UpliftBooster(
            outcome_budget=0.2, propensity_budget=0.05, effect_budget=0.1
        )
        self.assertEqual(model.outcome_budget, 0.2)
        self.assertEqual(model.propensity_budget, 0.05)
        self.assertEqual(model.effect_budget, 0.1)

    def test_fit_predict(self):
        model = UpliftBooster(effect_budget=0.1)
        model.fit(self.X, self.w, self.y)
        preds = model.predict(self.X)
        self.assertEqual(len(preds), 100)
        self.assertTrue(np.all(np.isfinite(preds)))

    def test_fit_pandas(self):
        X_df = pd.DataFrame(self.X, columns=[f"f{i}" for i in range(5)])
        model = UpliftBooster(effect_budget=0.1)
        model.fit(X_df, self.w, self.y)
        preds = model.predict(X_df)
        self.assertEqual(len(preds), 100)

    def test_serialization(self):
        model = UpliftBooster(effect_budget=0.1)
        model.fit(self.X, self.w, self.y)

        json_str = model.to_json()
        self.assertIsInstance(json_str, str)
        self.assertIn("outcome_model", json_str)
        self.assertIn("propensity_model", json_str)
        self.assertIn("effect_model", json_str)

        new_model = UpliftBooster.from_json(json_str)
        preds_old = model.predict(self.X)
        preds_new = new_model.predict(self.X)
        np.testing.assert_allclose(preds_old, preds_new, atol=1e-5)

    def test_invalid_inputs(self):
        model = UpliftBooster()
        with self.assertRaises(Exception):
            # w should be binary
            model.fit(self.X, np.random.randn(100), self.y)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pandas as pd
from perpetual import PerpetualBooster
from sklearn.datasets import make_classification, make_regression
from sklearn.model_selection import train_test_split


def explain_classification():
    print("\n--- Classification Explainability ---")
    # Generate synthetic data
    X, y = make_classification(
        n_samples=1000, n_features=10, n_informative=5, random_state=42
    )
    feature_names = [f"feature_{i}" for i in range(X.shape[1])]
    X_df = pd.DataFrame(X, columns=feature_names)

    X_train, X_test, y_train, y_test = train_test_split(
        X_df, y, test_size=0.2, random_state=42
    )

    # Train model
    model = PerpetualBooster(objective="LogLoss", budget=1.0)
    model.fit(X_train, y_train)

    # 1. Feature Importance
    print("\n1. Feature Importance (Gain):")
    importance = model.calculate_feature_importance(method="Gain", normalize=True)
    # Sort by importance
    sorted_importance = sorted(importance.items(), key=lambda x: x[1], reverse=True)
    for feat, imp in sorted_importance[:5]:  # Show top 5
        print(f"{feat}: {imp:.4f}")

    # 2. Partial Dependence
    print("\n2. Partial Dependence for top feature:")
    top_feature = sorted_importance[0][0]
    pd_values = model.partial_dependence(X_train, feature=top_feature, samples=10)
    print(f"Values for {top_feature}:")
    print(pd_values)

    # 3. Predict Contributions (SHAP-like)
    print("\n3. Prediction Contributions (SHAP-like) for first 2 samples:")
    # Calculate contributions
    contributions = model.predict_contributions(X_test.iloc[:2], method="Average")

    # contributions shape is (n_samples, n_features + 1), last column is bias
    bias = contributions[:, -1]
    feat_contribs = contributions[:, :-1]

    for i in range(2):
        print(f"\nSample {i}:")
        print(f"Bias: {bias[i]:.4f}")
        print("Top 3 contributing features:")
        # Get indices of top 3 absolute contributions
        top_idxs = np.argsort(np.abs(feat_contribs[i]))[-3:][::-1]
        for idx in top_idxs:
            print(f"{feature_names[idx]}: {feat_contribs[i, idx]:.4f}")

        prediction = model.predict(X_test.iloc[[i]])[0]
        # For LogLoss, contributions sum to the log-odds prediction
        sum_contribs = np.sum(feat_contribs[i]) + bias[i]
        print(f"Sum of contributions: {sum_contribs:.4f}")
        print(f"Model prediction: {prediction:.4f}")


def explain_regression():
    print("\n--- Regression Explainability ---")
    X, y = make_regression(
        n_samples=1000, n_features=10, n_informative=5, random_state=42
    )
    feature_names = [f"feature_{i}" for i in range(X.shape[1])]
    X_df = pd.DataFrame(X, columns=feature_names)

    X_train, X_test, y_train, y_test = train_test_split(
        X_df, y, test_size=0.2, random_state=42
    )

    model = PerpetualBooster(objective="SquaredLoss", budget=1.0)
    model.fit(X_train, y_train)

    print("\n1. Feature Importance (Cover):")
    importance = model.calculate_feature_importance(method="Cover", normalize=True)
    sorted_importance = sorted(importance.items(), key=lambda x: x[1], reverse=True)
    for feat, imp in sorted_importance[:5]:
        print(f"{feat}: {imp:.4f}")


if __name__ == "__main__":
    explain_classification()
    explain_regression()



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from perpetual import PerpetualBooster
from sklearn.model_selection import train_test_split


def test_load_v2_0_0_model():
    resource_dir = Path(__file__).parent.parent.parent.parent / "resources"
    model_path = resource_dir / "model_bc.json"

    if not model_path.exists():
        pytest.fail(
            f"Model artifact not found at {model_path}. Run scripts/make_resources.py to generate it."
        )

    # Load the model
    model = PerpetualBooster.load_booster(str(model_path))

    # Replicate Titanic data loading to match generation script
    df = pd.read_csv(resource_dir / "titanic.csv")
    X = df.drop(columns=["survived"])
    y = df["survived"]
    X["sex"] = pd.get_dummies(X["sex"], drop_first=True, dtype=float).to_numpy()
    X["adult_male"] = pd.get_dummies(
        X["adult_male"], drop_first=True, dtype=float
    ).to_numpy()
    X.drop(columns=["alive"], inplace=True)
    X["alone"] = pd.get_dummies(X["alone"], drop_first=True, dtype=float).to_numpy()
    cols = [
        "pclass",
        "sibsp",
        "parch",
        "embarked",
        "class",
        "who",
        "deck",
        "embark_town",
    ]
    X[cols] = X[cols].astype("category")

    data_train, data_test, y_train, _ = train_test_split(
        X, y, test_size=0.2, random_state=42
    )

    print(
        f"Current perpetual version: {PerpetualBooster.__module__.split('.')[0]}"
    )  # fallback if __version__ not directly available top-level
    import perpetual

    if hasattr(perpetual, "__version__"):
        print(f"Current perpetual version: {perpetual.__version__}")

    # Verify predictions
    preds = model.predict(data_test)

    # Assert predictions are valid (not NaN, correct shape)
    assert preds.shape[0] == data_test.shape[0]
    assert not np.isnan(preds).any()

    # Check objective
    assert model.objective == "LogLoss"

    # Compare with saved backward-compatibility predictions
    preds_path = resource_dir / "model_bc_preds.csv"
    if preds_path.exists():
        expected_preds = (
            pd.read_csv(preds_path, header=None).squeeze("columns").to_numpy()
        )
        if model.objective == "LogLoss":
            np.testing.assert_array_equal(np.rint(preds), np.rint(expected_preds))

            # Compare probabilities
            probs_path = resource_dir / "model_bc_probs.csv"
            expected_probs = pd.read_csv(probs_path, header=None).to_numpy()
            preds_probs = model.predict_proba(data_test)
            np.testing.assert_allclose(preds_probs, expected_probs, rtol=1e-5)
        else:
            np.testing.assert_allclose(preds, expected_preds, rtol=1e-5)
    else:
        pytest.fail(f"Prediction artifact not found at {preds_path}")



"""Perpetual — a self-generalizing Gradient Boosting Machine.

Provides:
  - :class:`PerpetualBooster` — single-output gradient booster.
  - :class:`PerpetualClassifier`, :class:`PerpetualRegressor`, :class:`PerpetualRanker` — scikit-learn wrappers.
  - :class:`MultiOutputBooster` — multi-target booster (via internal import).
  - :class:`UpliftBooster` — R-Learner uplift estimator.
  - :class:`IVBooster` — Instrumental Variable (BoostIV) estimator.
  - :class:`SLearner`, :class:`TLearner`, :class:`XLearner`, :class:`DRLearner` — meta-learner CATE estimators.
  - :class:`DMLEstimator` — Double Machine Learning (DML) estimator with cross-fitting.
  - :class:`FairClassifier` — fairness-aware classifier (Demographic Parity / Equalized Odds).
  - :class:`PolicyLearner` — IPW / AIPW policy learning.
  - :class:`PerpetualRiskEngine` — adverse-action reason-code generator.
  - :mod:`causal_metrics` — AUUC, Qini coefficient, cumulative gain curves.
"""

from __future__ import annotations

from perpetual.booster import (
    CalibrationMethod,
    MultiOutputBooster,
    PerpetualBooster,
    compute_calibration_curve,
    expected_calibration_error,
)
from perpetual.causal_metrics import (
    auuc,
    cumulative_gain_curve,
    qini_coefficient,
    qini_curve,
)
from perpetual.dml import DMLEstimator
from perpetual.fairness import FairClassifier
from perpetual.iv import IVBooster
from perpetual.meta_learners import DRLearner, SLearner, TLearner, XLearner
from perpetual.policy import PolicyLearner
from perpetual.risk import PerpetualRiskEngine
from perpetual.sklearn import PerpetualClassifier, PerpetualRanker, PerpetualRegressor
from perpetual.uplift import UpliftBooster

__all__ = [
    "PerpetualBooster",
    "MultiOutputBooster",
    "CalibrationMethod",
    "compute_calibration_curve",
    "expected_calibration_error",
    "PerpetualClassifier",
    "PerpetualRegressor",
    "PerpetualRanker",
    "UpliftBooster",
    "IVBooster",
    "PerpetualRiskEngine",
    "SLearner",
    "TLearner",
    "XLearner",
    "DRLearner",
    "DMLEstimator",
    "FairClassifier",
    "PolicyLearner",
    "auuc",
    "cumulative_gain_curve",
    "qini_curve",
    "qini_coefficient",
]



import argparse
import json
import re
import warnings
from functools import partial
from pathlib import Path
from time import process_time, time

import numpy as np
import optuna
import pandas as pd
import requests
from catboost import CatBoostClassifier, CatBoostRegressor
from perpetual import PerpetualBooster
from sklearn.datasets import fetch_openml
from sklearn.metrics import log_loss, mean_squared_error, roc_auc_score
from sklearn.model_selection import KFold, StratifiedKFold, train_test_split

# python package-python/examples/benchmark_comparison.py
# python package-python/examples/benchmark_comparison.py --no-optuna
# python package-python/examples/benchmark_comparison.py --skip-catboost
# python package-python/examples/benchmark_comparison.py --catboost-only
# python package-python/examples/benchmark_comparison.py --catboost-only --catboost-results-path tmp/catboost_results.csv

# https://www.openml.org/search?type=study&study_type=task&id=457&sort=tasks_included


"""
All PB (tuned) tabarena performance (v2.1.0):
                                   dataset problem_type   metric     method  metric_error  rank  total_methods dataset_id performance
                                       MIC   multiclass log_loss PB (tuned)      1.238202  18.0             18 46980
                             seismic-bumps       binary  roc_auc PB (tuned)      0.258542  18.0             18 46956      3-2
           taiwanese_bankruptcy_prediction       binary  roc_auc PB (tuned)      0.160871  18.0             18 46962      4-1
     students_dropout_and_academic_success   multiclass log_loss PB (tuned)      1.125762  18.0             18
                   Is-this-a-good-customer       binary  roc_auc PB (tuned)      0.279860  17.0             18
          blood-transfusion-service-center       binary  roc_auc PB (tuned)      0.277322  17.0             18 46913
                                     heloc       binary  roc_auc PB (tuned)      0.212887  17.0             18
                                    anneal   multiclass log_loss PB (tuned)      0.067830  17.0             18 46906
                                     churn       binary  roc_auc PB (tuned)      0.136474  17.0             18
                                  diabetes       binary  roc_auc PB (tuned)      0.175348  17.0             18 46921
                        kddcup09_appetency       binary  roc_auc PB (tuned)      0.369296  17.0             17
               coil2000_insurance_policies       binary  roc_auc PB (tuned)      0.267649  17.0             18
                                    splice   multiclass log_loss PB (tuned)      0.308573  17.0             18
                             Diabetes130US       binary  roc_auc PB (tuned)      0.411460  17.0             17
                    Amazon_employee_access       binary  roc_auc PB (tuned)      0.186921  17.0             17
                              Fitness_Club       binary  roc_auc PB (tuned)      0.194381  16.0             18
                                APSFailure       binary  roc_auc PB (tuned)      0.013972  16.0             17
HR_Analytics_Job_Change_of_Data_Scientists       binary  roc_auc PB (tuned)      0.210702  16.0             17
                     E-CommereShippingData       binary  roc_auc PB (tuned)      0.263426  16.0             18
                             hiva_agnostic   multiclass log_loss PB (tuned)      0.309314  16.0             17
                          website_phishing   multiclass log_loss PB (tuned)      0.294513  16.0             18
                        airfoil_self_noise   regression     rmse PB (tuned)      2.417587  16.0             18
                        QSAR_fish_toxicity   regression     rmse PB (tuned)      0.899660  15.0             18 46954
                 online_shoppers_intention       binary  roc_auc PB (tuned)      0.069808  15.0             18
                               qsar-biodeg       binary  roc_auc PB (tuned)      0.075866  15.0             18
               credit_card_clients_default       binary  roc_auc PB (tuned)      0.225365  15.0             17
                                    SDSS17   multiclass log_loss PB (tuned)      0.131018  15.0             17 46955
                               QSAR-TID-11   regression     rmse PB (tuned)      0.813204  15.0             17
                            bank-marketing       binary  roc_auc PB (tuned)      0.248461  15.0             17
                                  credit-g       binary  roc_auc PB (tuned)      0.218830  14.0             18 46918
                      maternal_health_risk   multiclass log_loss PB (tuned)      0.564076  14.0             18
                          GiveMeSomeCredit       binary  roc_auc PB (tuned)      0.142017  14.0             17 46929
                       Bank_Customer_Churn       binary  roc_auc PB (tuned)      0.139907  14.0             18
                        Marketing_Campaign       binary  roc_auc PB (tuned)      0.096731  14.0             18
                                       jm1       binary  roc_auc PB (tuned)      0.254839  13.0             18
                              wine_quality   regression     rmse PB (tuned)      0.629992  13.0             18
             concrete_compressive_strength   regression     rmse PB (tuned)      4.928130  13.0             18
                    physiochemical_protein   regression     rmse PB (tuned)      3.562011  13.0             17
               polish_companies_bankruptcy       binary  roc_auc PB (tuned)      0.063708  13.0             18
             healthcare_insurance_expenses   regression     rmse PB (tuned)   4625.943316  12.0             18
                             miami_housing   regression     rmse PB (tuned)  85758.029216  11.0             18
     hazelnut-spread-contaminant-detection       binary  roc_auc PB (tuned)      0.022465  11.0             18
          Another-Dataset-on-used-Fiat-500   regression     rmse PB (tuned)    746.753007  11.0             18
                         superconductivity   regression     rmse PB (tuned)      9.575404  11.0             17
          customer_satisfaction_in_airline       binary  roc_auc PB (tuned)      0.005844  11.0             17 46920
                        Food_Delivery_Time   regression     rmse PB (tuned)      7.488482  10.0             17
                                  diamonds   regression     rmse PB (tuned)    531.637650  10.0             17
                                    houses   regression     rmse PB (tuned)      0.217172  10.0             17
                              NATICUSdroid       binary  roc_auc PB (tuned)      0.014906   9.0             18
          in_vehicle_coupon_recommendation       binary  roc_auc PB (tuned)      0.166644   9.0             18
                               Bioresponse       binary  roc_auc PB (tuned)      0.123246   2.0             17
"""


# Dataset IDs with the most number of rows
# TASK_IDS = [46955, 46920, 46929]
# Dataset IDs with the least number of rows (max 1000 rows)
# TASK_IDS = [46913, 46921, 46906, 46954, 46918]
# All 11 datasets
TASK_IDS = [46913, 46921, 46906, 46954, 46918, 46955, 46920, 46929, 46980, 46956, 46962]

cb_iterations = 100
n_trials = 100
PERPETUAL_BUDGET = 2.0
use_optuna = True  # Set to False to use default CatBoost parameters (skip Optuna)
skip_catboost = False  # Set to True to skip CatBoost completely
n_splits = 3


def resolve_task(task_id):
    """
    Fetches the source dataset ID and target feature for a given OpenML task ID.
    """
    url = f"https://www.openml.org/api/v1/xml/task/{task_id}"
    try:
        response = requests.get(url, timeout=10)
        if response.status_code == 200:
            text = response.text
            did_match = re.search(r"<oml:data_set_id>(\d+)</oml:data_set_id>", text)
            target_match = re.search(
                r"<oml:target_feature>(.*?)</oml:target_feature>", text
            )

            did = int(did_match.group(1)) if did_match else task_id
            target = target_match.group(1) if target_match else None
            return did, target
    except Exception as e:
        print(f"Warning: Could not resolve task {task_id}: {e}")
    return task_id, None


def prepare_data(task_id, seed):
    # Resolve Task ID to Dataset ID and target feature
    data_id, target = resolve_task(task_id)

    print(f"Loading Dataset ID {data_id} (from Task {task_id})...")
    try:
        # Fetch the dataset
        data = fetch_openml(
            data_id=data_id, return_X_y=True, as_frame=True, parser="auto"
        )
        X, y = data
        print(f"Data shape: {X.shape}")
    except Exception as e:
        print(f"Error fetching data for data_id {data_id}: {e}")
        raise

    # If y is a DataFrame (multi-target) or if we have a specific target name
    if isinstance(y, pd.DataFrame):
        if target and target in y.columns:
            y = y[target]
        else:
            y = y.iloc[:, 0]

    # Simple task type detection
    if (
        y.dtype == "object"
        or y.dtype.name == "category"
        or len(np.unique(y.dropna())) < 20
    ):
        n_classes = len(np.unique(y.dropna()))
        if n_classes == 2:
            task_type = "binary"
            scoring = "roc_auc"
            metric_function = roc_auc_score
        else:
            task_type = "multiclass"
            scoring = "neg_log_loss"
            metric_function = log_loss
    else:
        task_type = "regression"
        scoring = "neg_mean_squared_error"
        metric_function = mean_squared_error

    X_train, X_test, y_train, y_test = train_test_split(
        X,
        y,
        test_size=0.20,
        random_state=seed,
        stratify=y if task_type in {"binary", "multiclass"} else None,
    )

    return (X_train, X_test, y_train, y_test, task_type, scoring, metric_function)


def build_cv(task_type, seed):
    if task_type in {"binary", "multiclass"}:
        return StratifiedKFold(n_splits=n_splits, shuffle=True, random_state=seed)
    return KFold(n_splits=n_splits, shuffle=True, random_state=seed)


def score_predictions(task_type, y_true, y_pred):
    if task_type == "binary":
        return roc_auc_score(y_true, y_pred)
    if task_type == "multiclass":
        return -log_loss(y_true, y_pred)
    return -mean_squared_error(y_true, y_pred)


def evaluate_predictions(task_type, y_true, y_pred):
    if task_type == "binary":
        return roc_auc_score(y_true, y_pred)
    if task_type == "multiclass":
        return log_loss(y_true, y_pred)
    return mean_squared_error(y_true, y_pred)


def predict_for_metric(model, X_data, task_type):
    if task_type == "binary":
        return model.predict_proba(X_data)[:, 1]
    if task_type == "multiclass":
        return model.predict_proba(X_data)
    return model.predict(X_data)


def objective_catboost(trial, X_train, y_train, task_type, seed):
    params = {
        "iterations": cb_iterations,
        "learning_rate": trial.suggest_float("learning_rate", 0.001, 0.5, log=True),
        "depth": trial.suggest_int("depth", 4, 10),
        "l2_leaf_reg": trial.suggest_float("l2_leaf_reg", 1e-6, 10.0, log=True),
        "random_strength": trial.suggest_float("random_strength", 1e-6, 10.0, log=True),
        "bagging_temperature": trial.suggest_float("bagging_temperature", 0.0, 1.0),
        "od_type": "Iter",
        "od_wait": 20,
        "verbose": False,
        "random_seed": seed,
        "allow_writing_files": False,
    }

    # Identify cat features
    cat_features = X_train.select_dtypes(
        include=["object", "category"]
    ).columns.tolist()

    cv = build_cv(task_type, seed)

    scores = []
    for train_idx, val_idx in cv.split(X_train, y_train):
        X_tr, X_val = X_train.iloc[train_idx], X_train.iloc[val_idx]
        y_tr, y_val = y_train.iloc[train_idx], y_train.iloc[val_idx]

        if task_type in {"binary", "multiclass"}:
            model = CatBoostClassifier(**params)
        else:
            model = CatBoostRegressor(**params)

        model.fit(
            X_tr,
            y_tr,
            eval_set=(X_val, y_val),
            early_stopping_rounds=20,
            cat_features=cat_features,
        )

        preds = predict_for_metric(model, X_val, task_type)
        score = score_predictions(task_type, y_val, preds)
        scores.append(score)

    return np.mean(scores)


def run_catboost(X_train, y_train, X_test, y_test, task_type, seed):
    start_cpu = process_time()
    start_wall = time()

    if use_optuna:
        sampler = optuna.samplers.TPESampler(seed=seed)
        study = optuna.create_study(direction="maximize", sampler=sampler)

        obj = partial(
            objective_catboost,
            X_train=X_train,
            y_train=y_train,
            task_type=task_type,
            seed=seed,
        )

        study.optimize(obj, n_trials=n_trials)
        best_params = study.best_params
        best_params["iterations"] = cb_iterations
    else:
        best_params = {}

    best_params["verbose"] = False
    best_params["random_seed"] = seed
    best_params["allow_writing_files"] = False

    cv = build_cv(task_type, seed)

    ensemble_preds = []
    cat_features = X_train.select_dtypes(
        include=["object", "category"]
    ).columns.tolist()

    for train_idx, val_idx in cv.split(X_train, y_train):
        X_tr, y_tr = X_train.iloc[train_idx], y_train.iloc[train_idx]
        if task_type in {"binary", "multiclass"}:
            model = CatBoostClassifier(**best_params)
        else:
            model = CatBoostRegressor(**best_params)

        model.fit(X_tr, y_tr, cat_features=cat_features)

        ensemble_preds.append(predict_for_metric(model, X_test, task_type))

    y_pred = np.mean(ensemble_preds, axis=0)
    stop_cpu = process_time()
    stop_wall = time()

    metric = evaluate_predictions(task_type, y_test, y_pred)

    best_cv_score = study.best_value if use_optuna else None

    return {
        "metric": metric,
        "cpu_time": stop_cpu - start_cpu,
        "wall_time": stop_wall - start_wall,
        "best_cv_score": best_cv_score,
        "best_params": dict(best_params),
    }


def run_perpetual(X_train, y_train, X_test, y_test, task_type, seed):
    objective = "LogLoss" if task_type in {"binary", "multiclass"} else "SquaredLoss"
    model = PerpetualBooster(
        objective=objective,
        budget=PERPETUAL_BUDGET,
        max_cat=1000,
        seed=seed,
        iteration_limit=10000,
    )

    start_cpu = process_time()
    start_wall = time()
    model.fit(X_train, y_train)
    stop_cpu = process_time()
    stop_wall = time()

    y_pred = predict_for_metric(model, X_test, task_type)
    metric = evaluate_predictions(task_type, y_test, y_pred)

    return metric, stop_cpu - start_cpu, stop_wall - start_wall


if __name__ == "__main__":
    warnings.filterwarnings("ignore")
    optuna.logging.set_verbosity(optuna.logging.WARNING)

    parser = argparse.ArgumentParser(
        description=(
            "Benchmark PB vs CatBoost. Use --no-optuna to skip CatBoost tuning "
            "or --catboost-only to tune CatBoost and save its results without "
            "running PerpetualBooster."
        )
    )
    parser.add_argument(
        "--no-optuna",
        action="store_true",
        help="Skip CatBoost Optuna optimization and use default parameters",
    )
    parser.add_argument(
        "--skip-catboost",
        action="store_true",
        help="Skip CatBoost completely",
    )
    parser.add_argument(
        "--catboost-only",
        action="store_true",
        help=(
            "Run only CatBoost (with Optuna unless --no-optuna is set), skip "
            "PerpetualBooster, and save CatBoost results to CSV"
        ),
    )
    parser.add_argument(
        "--catboost-results-path",
        default=None,
        help=(
            "Optional CSV path for saved CatBoost results. Defaults to "
            "catboost_results.csv when --catboost-only is used. When used with --skip-catboost, reads results from this file to compare."
        ),
    )
    args = parser.parse_args()

    # Initialize execution flags with default values.
    use_optuna = True
    skip_catboost = False
    skip_perpetual = False

    if args.catboost_only and args.skip_catboost:
        parser.error("--catboost-only cannot be combined with --skip-catboost")

    if args.no_optuna:
        use_optuna = False
        print("Optuna optimization skipped. Using default CatBoost parameters.")

    if args.skip_catboost:
        skip_catboost = True
        print("CatBoost evaluation skipped completely.")

    catboost_results_path = None
    if args.catboost_only:
        skip_perpetual = True
        catboost_results_path = Path(
            args.catboost_results_path or "catboost_results.csv"
        )
        print(
            "Running CatBoost-only mode. PerpetualBooster evaluation skipped; "
            f"results will be saved to {catboost_results_path}."
        )
    elif args.catboost_results_path:
        catboost_results_path = Path(args.catboost_results_path)

    catboost_saved_rows = []

    loaded_catboost_results = {}
    if skip_catboost and catboost_results_path and catboost_results_path.exists():
        cb_df = pd.read_csv(catboost_results_path)
        for _, row in cb_df.iterrows():
            loaded_catboost_results[(int(row["task_id"]), int(row["seed"]))] = row[
                "metric"
            ]

    all_results = []

    for data_id in TASK_IDS:
        print(f"\nEvaluating Task ID: {data_id}")
        task_results = {"data_id": data_id}

        # Methodology: 5 seeds
        seeds = range(5)

        for seed in seeds:
            try:
                (
                    X_train,
                    X_test,
                    y_train,
                    y_test,
                    task_type,
                    scoring,
                    metric_function,
                ) = prepare_data(data_id, seed)

                # CatBoost
                if not skip_catboost:
                    cb_result = run_catboost(
                        X_train, y_train, X_test, y_test, task_type, seed
                    )
                    print(
                        "Seed "
                        f"{seed} - CatBoost: {cb_result['metric']:.4f}, CPU: "
                        f"{cb_result['cpu_time']:.2f}s, Wall: "
                        f"{cb_result['wall_time']:.2f}s"
                    )
                    if "CatBoost" not in task_results:
                        task_results["CatBoost"] = []
                    task_results["CatBoost"].append(cb_result["metric"])

                    if catboost_results_path is not None and not skip_catboost:
                        catboost_saved_rows.append(
                            {
                                "task_id": data_id,
                                "seed": seed,
                                "task_type": task_type,
                                "metric": cb_result["metric"],
                                "cpu_time": cb_result["cpu_time"],
                                "wall_time": cb_result["wall_time"],
                                "use_optuna": use_optuna,
                                "best_cv_score": cb_result["best_cv_score"],
                                "best_params": json.dumps(
                                    cb_result["best_params"], sort_keys=True
                                ),
                            }
                        )
                elif skip_catboost and (data_id, seed) in loaded_catboost_results:
                    cb_metric = loaded_catboost_results[(data_id, seed)]
                    if "CatBoost" not in task_results:
                        task_results["CatBoost"] = []
                    task_results["CatBoost"].append(cb_metric)

                # PerpetualBooster fixed-budget benchmark run.
                if not skip_perpetual:
                    pb_metric, pb_cpu, pb_wall = run_perpetual(
                        X_train, y_train, X_test, y_test, task_type, seed
                    )
                    print(
                        "Seed "
                        f"{seed} - Perpetual budget={PERPETUAL_BUDGET:.1f}: "
                        f"Test={pb_metric:.4f}, CPU: "
                        f"{pb_cpu:.2f}s, Wall: {pb_wall:.2f}s"
                    )

                    if "PB_budget" not in task_results:
                        task_results["PB_budget"] = []
                    task_results["PB_budget"].append(PERPETUAL_BUDGET)

                    if "PB_test" not in task_results:
                        task_results["PB_test"] = []
                    task_results["PB_test"].append(pb_metric)

            except Exception as e:
                print(f"Error evaluating Dataset ID {data_id} with seed {seed}: {e}")
                continue

        all_results.append(task_results)

    # Print final summary table
    print("\n" + "=" * 50)
    print("FINAL SUMMARY (Averaged over 5 seeds)")
    print("=" * 50)
    for res in all_results:
        summary = f"Dataset {res['data_id']}: "
        if "CatBoost" in res:
            summary += f"CatBoost={np.mean(res['CatBoost']):.4f}, "

        if "PB_test" in res:
            wins = None
            if "CatBoost" in res:
                wins = sum(pb > cb for pb, cb in zip(res["PB_test"], res["CatBoost"]))
            summary += f"PB(test)={np.mean(res['PB_test']):.4f}, PB(budget)=2.0"
            if wins is not None:
                summary += f", PB wins={wins}/5"

        print(summary)

    if catboost_results_path is not None and catboost_saved_rows:
        catboost_results_path.parent.mkdir(parents=True, exist_ok=True)
        pd.DataFrame(catboost_saved_rows).to_csv(catboost_results_path, index=False)
        print(f"Saved CatBoost results to {catboost_results_path}")

"""
==================================================
FINAL SUMMARY (Averaged over 5 seeds)
==================================================
Catboost with default parameters (no Optuna):
Dataset 46956: CatBoost=0.7690, PB_0.5=0.7269, PB_1.0=0.7531, PB_1.5=0.7525, PB_2.0=0.7530, PB_2.5=0.7563

((self.cfg.budget.max(1.0) * cols.max(1) as f32) * 10.0).round().clamp(256.0, ITER_LIMIT as f32) as usize
Dataset 46956:                  PB_0.5=0.7491, PB_1.0=0.7608, PB_1.5=0.7523, PB_2.0=0.7547, PB_2.5=0.7707

# DO NOT SET ITERATION_LIMIT
StructuralStopState - budget < 2.0:
Dataset 46956:                  PB_0.5=0.7491, PB_1.0=0.7608, PB_1.5=0.7523, PB_2.0=0.7624, PB_2.5=0.7713

# DO NOT SET ITERATION_LIMIT
StructuralStopState - budget < 0.5: 
Dataset 46956:                  PB_0.5=0.7521, PB_1.0=0.7593, PB_1.5=0.7561, PB_2.0=0.7624, PB_2.5=0.7713


Evaluating Task ID: 46956
Loading Dataset ID 46956 (from Task 46956)...
Data shape: (2584, 15)
Seed 0 - CatBoost: 0.7339, CPU: 5193.06s, Wall: 2657.54s
Seed 0 - Perpetual CV (budget=0.5): 0.7519
Seed 0 - Perpetual CV (budget=1.0): 0.7721
Seed 0 - Perpetual CV (budget=1.5): 0.7736
Seed 0 - Perpetual CV (budget=2.0): 0.7720
Seed 0 - Perpetual CV (budget=2.5): 0.7708
Seed 0 - Perpetual best budget=1.5: CV=0.7736, Test=0.6891, CPU: 17.52s, Wall: 6.78s
Loading Dataset ID 46956 (from Task 46956)...
Data shape: (2584, 15)
Seed 1 - CatBoost: 0.7813, CPU: 3778.50s, Wall: 2313.68s
Seed 1 - Perpetual CV (budget=0.5): 0.7270
Seed 1 - Perpetual CV (budget=1.0): 0.7411
Seed 1 - Perpetual CV (budget=1.5): 0.7394
Seed 1 - Perpetual CV (budget=2.0): 0.7500
Seed 1 - Perpetual CV (budget=2.5): 0.7457
Seed 1 - Perpetual best budget=2.0: CV=0.7500, Test=0.7895, CPU: 18.23s, Wall: 6.19s
Loading Dataset ID 46956 (from Task 46956)...
Data shape: (2584, 15)
Seed 2 - CatBoost: 0.7926, CPU: 3518.80s, Wall: 2120.20s
Seed 2 - Perpetual CV (budget=0.5): 0.7430
Seed 2 - Perpetual CV (budget=1.0): 0.7671
Seed 2 - Perpetual CV (budget=1.5): 0.7611
Seed 2 - Perpetual CV (budget=2.0): 0.7641
Seed 2 - Perpetual CV (budget=2.5): 0.7719
Seed 2 - Perpetual best budget=2.5: CV=0.7719, Test=0.8183, CPU: 19.28s, Wall: 6.75s
Loading Dataset ID 46956 (from Task 46956)...
Data shape: (2584, 15)
Seed 3 - CatBoost: 0.7282, CPU: 3405.52s, Wall: 1843.29s
Seed 3 - Perpetual CV (budget=0.5): 0.7425
Seed 3 - Perpetual CV (budget=1.0): 0.7713
Seed 3 - Perpetual CV (budget=1.5): 0.7685
Seed 3 - Perpetual CV (budget=2.0): 0.7711
Seed 3 - Perpetual CV (budget=2.5): 0.7689
Seed 3 - Perpetual best budget=1.0: CV=0.7713, Test=0.7717, CPU: 19.77s, Wall: 6.93s
Loading Dataset ID 46956 (from Task 46956)...
Data shape: (2584, 15)
Seed 4 - CatBoost: 0.7539, CPU: 2376.50s, Wall: 1441.93s
Seed 4 - Perpetual CV (budget=0.5): 0.7460
Seed 4 - Perpetual CV (budget=1.0): 0.7610
Seed 4 - Perpetual CV (budget=1.5): 0.7658
Seed 4 - Perpetual CV (budget=2.0): 0.7744
Seed 4 - Perpetual CV (budget=2.5): 0.7731
Seed 4 - Perpetual best budget=2.0: CV=0.7744, Test=0.7538, CPU: 19.03s, Wall: 5.77s

==================================================
FINAL SUMMARY (Averaged over 5 seeds)
==================================================
Dataset 46956: CatBoost=0.7580, PB(test)=0.7645, PB(CV score)=0.7683, PB(best budget avg)=1.80


Evaluating Task ID: 46962
Loading Dataset ID 46962 (from Task 46962)...
Data shape: (6819, 94)
Seed 0 - CatBoost: 0.9294, CPU: 26482.53s, Wall: 4417.47s
Seed 0 - Perpetual CV (budget=0.5): 0.9315
Seed 0 - Perpetual CV (budget=1.0): 0.9507
Seed 0 - Perpetual CV (budget=1.5): 0.9503
Seed 0 - Perpetual CV (budget=2.0): 0.9501
Seed 0 - Perpetual CV (budget=2.5): 0.9484
Seed 0 - Perpetual best budget=1.0: CV=0.9507, Test=0.9338, CPU: 13021.89s, Wall: 2424.28s
Loading Dataset ID 46962 (from Task 46962)...
Data shape: (6819, 94)
Seed 1 - CatBoost: 0.9528, CPU: 8262.08s, Wall: 1073.73s
Seed 1 - Perpetual CV (budget=0.5): 0.9317
Seed 1 - Perpetual CV (budget=1.0): 0.9382
Seed 1 - Perpetual CV (budget=1.5): 0.9379
Seed 1 - Perpetual CV (budget=2.0): 0.9425
Seed 1 - Perpetual CV (budget=2.5): 0.9391
Seed 1 - Perpetual best budget=2.0: CV=0.9425, Test=0.9535, CPU: 7514.67s, Wall: 1250.49s
Loading Dataset ID 46962 (from Task 46962)...
Data shape: (6819, 94)
Seed 2 - CatBoost: 0.9546, CPU: 8852.00s, Wall: 1229.24s
Seed 2 - Perpetual CV (budget=0.5): 0.9225
Seed 2 - Perpetual CV (budget=1.0): 0.9381
Seed 2 - Perpetual CV (budget=1.5): 0.9387
Seed 2 - Perpetual CV (budget=2.0): 0.9389
Seed 2 - Perpetual CV (budget=2.5): 0.9378
Seed 2 - Perpetual CV (budget=2.5): 0.9378
Seed 2 - Perpetual best budget=2.0: CV=0.9389, Test=0.9590, CPU: 9437.67s, Wall: 1530.59s
Loading Dataset ID 46962 (from Task 46962)...
Data shape: (6819, 94)
Seed 3 - CatBoost: 0.9359, CPU: 13050.34s, Wall: 1737.22s
Seed 3 - Perpetual CV (budget=0.5): 0.9196
Seed 3 - Perpetual CV (budget=1.0): 0.9444
Seed 3 - Perpetual CV (budget=1.5): 0.9426
Seed 3 - Perpetual CV (budget=2.0): 0.9445
Seed 3 - Perpetual CV (budget=2.5): 0.9436
Seed 3 - Perpetual best budget=2.0: CV=0.9445, Test=0.9298, CPU: 10488.19s, Wall: 1712.49s
Loading Dataset ID 46962 (from Task 46962)...
Data shape: (6819, 94)
Seed 4 - CatBoost: 0.9444, CPU: 67732.55s, Wall: 10211.18s
Seed 4 - Perpetual CV (budget=0.5): 0.9271
Seed 4 - Perpetual CV (budget=1.0): 0.9441
Seed 4 - Perpetual CV (budget=1.5): 0.9381
Seed 4 - Perpetual CV (budget=2.0): 0.9379
Seed 4 - Perpetual CV (budget=2.5): 0.9368
Seed 4 - Perpetual best budget=1.0: CV=0.9441, Test=0.9458, CPU: 9515.03s, Wall: 2417.26s

==================================================
FINAL SUMMARY (Averaged over 5 seeds)
==================================================
Dataset 46962: CatBoost=0.9434, PB(test)=0.9444, PB(CV score)=0.9441, PB(best budget avg)=1.60



row subsampling and max_cat=1000:

Evaluating Task ID: 46955
Loading Dataset ID 46955 (from Task 46955)...
Data shape: (78053, 11)
Seed 0 - CatBoost: 0.0769, CPU: 16382.72s, Wall: 3170.04s
Seed 0 - Perpetual CV (budget=0.5): -0.1410
Seed 0 - Perpetual CV (budget=1.0): -0.1125
Seed 0 - Perpetual CV (budget=1.5): -0.1413
Seed 0 - Perpetual CV (budget=2.0): -0.3378
Seed 0 - Perpetual CV (budget=2.5): -0.8828
Seed 0 - Perpetual best budget=1.0: CV=-0.1125, Test=0.1134, CPU: 152.23s, Wall: 50.51s


no row subsampling and max_cat=9000:

Evaluating Task ID: 46955
Loading Dataset ID 46955 (from Task 46955)...
Data shape: (78053, 11)
Seed 0 - CatBoost: 0.0769, CPU: 23710.47s, Wall: 4679.17s
Seed 0 - Perpetual CV (budget=0.5): -0.3614
Seed 0 - Perpetual CV (budget=1.0): -0.2878
Seed 0 - Perpetual CV (budget=1.5): -0.2289
Seed 0 - Perpetual CV (budget=2.0): -0.1945
Seed 0 - Perpetual CV (budget=2.5): -0.1684
Seed 0 - Perpetual best budget=2.5: CV=-0.1684, Test=0.1782, CPU: 1871.59s, Wall: 1351.24s
Loading Dataset ID 46955 (from Task 46955)...
Data shape: (78053, 11)


no row subsampling and max_cat=1000:

Evaluating Task ID: 46955
Loading Dataset ID 46955 (from Task 46955)...
Data shape: (78053, 11)
Seed 0 - CatBoost: 0.0769, CPU: 20905.67s, Wall: 3693.78s
Seed 0 - Perpetual CV (budget=0.5): -0.1410
Seed 0 - Perpetual CV (budget=1.0): -0.1190
Seed 0 - Perpetual CV (budget=1.5): -0.1037
Seed 0 - Perpetual CV (budget=2.0): -0.1687
Seed 0 - Perpetual CV (budget=2.5): -0.1788
Seed 0 - Perpetual best budget=1.5: CV=-0.1037, Test=0.1558, CPU: 325.73s, Wall: 167.95s
"""



"""Boosted Instrumental Variable (BoostIV) estimator.

Implements a two-stage least-squares approach using gradient boosting to
estimate causal effects in the presence of endogeneity.
"""

from typing import Any, Dict, Iterable, Optional, Union

import numpy as np
from typing_extensions import Self

from perpetual.perpetual import IVBooster as CrateIVBooster
from perpetual.utils import convert_input_array, convert_input_frame


class IVBooster:
    """Two-stage instrumental-variable estimator powered by gradient boosting.

    Stage 1 regresses the treatment on the instruments, and Stage 2 regresses
    the outcome on the predicted treatment and covariates.  Both stages are
    fitted using Perpetual's self-generalizing boosting.
    """

    def __init__(
        self,
        treatment_objective: str = "SquaredLoss",
        outcome_objective: str = "SquaredLoss",
        stage1_budget: float = 0.5,
        stage2_budget: float = 0.5,
        num_threads: Optional[int] = None,
        monotone_constraints: Union[Dict[Any, int], None] = None,
        force_children_to_bound_parent: bool = False,
        missing: float = np.nan,
        allow_missing_splits: bool = True,
        create_missing_branch: bool = False,
        terminate_missing_features: Optional[Iterable[Any]] = None,
        missing_node_treatment: str = "None",
        log_iterations: int = 0,
        reset: Optional[bool] = None,
        categorical_features: Union[Iterable[int], Iterable[str], str, None] = "auto",
        timeout: Optional[float] = None,
        iteration_limit: Optional[int] = None,
        memory_limit: Optional[float] = None,
        stopping_rounds: Optional[int] = None,
        max_bin: int = 256,
        max_cat: int = 1000,
        interaction_constraints: Optional[list[list[int]]] = None,
        seed: int = 0,
        **kwargs,
    ):
        """
        Boosted Instrumental Variable (BoostIV) Estimator.

        Implements a 2-Stage Least Squares (2SLS) approach using Gradient Boosting.

        Parameters
        ----------
        treatment_objective : str, default="SquaredLoss"
            Objective for Stage 1 (Treatment Model). e.g., ``"SquaredLoss"`` or ``"LogLoss"``.
        outcome_objective : str, default="SquaredLoss"
            Objective for Stage 2 (Outcome Model). e.g., ``"SquaredLoss"``.
        stage1_budget : float, default=0.5
            Fitting budget for Stage 1. Higher values allow more boosting rounds.
        stage2_budget : float, default=0.5
            Fitting budget for Stage 2. Higher values allow more boosting rounds.
        num_threads : int, optional
            Number of threads to use during training and prediction.
        monotone_constraints : dict, optional
            Constraints mapping feature indices/names to -1, 1, or 0.
        force_children_to_bound_parent : bool, default=False
            Whether to restrict children nodes to be within the parent's range.
        missing : float, default=np.nan
            Value to consider as missing data.
        allow_missing_splits : bool, default=True
            Whether to allow splits that separate missing from non-missing values.
        create_missing_branch : bool, default=False
            Whether to create a separate branch for missing values (ternary trees).
        terminate_missing_features : iterable, optional
            Features for which missing branches are always terminated when
            ``create_missing_branch`` is True.
        missing_node_treatment : str, default="None"
            How to handle weights for missing nodes. Options: ``"None"``,
            ``"AssignToParent"``, ``"AverageLeafWeight"``, ``"AverageNodeWeight"``.
        log_iterations : int, default=0
            Logging frequency (every N iterations). 0 disables logging.
        **kwargs
            Objective-specific parameters passed to the booster core:

            - **quantile** (float): The target quantile for ``"QuantileLoss"`` and
              ``"AdaptiveHuberLoss"``. Must be between 0 and 1. Default is 0.5.
            - **delta** (float): The threshold parameter for ``"HuberLoss"``. Default is 1.0.
            - **c** (float): The scale parameter for ``"FairLoss"``. Default is 1.0.
            - **p** (float): The power parameter for ``"TweedieLoss"``. Must be between 1 and 2.
              Default is 1.5.
        reset : bool, optional
            Whether to reset the model or continue training on subsequent fits.
        categorical_features : iterable or str, default="auto"
            Feature indices or names to treat as categorical.
        timeout : float, optional
            Time limit for fitting in seconds.
        iteration_limit : int, optional
            Maximum number of boosting iterations.
        memory_limit : float, optional
            Memory limit for training in GB.
        stopping_rounds : int, optional
            Number of rounds without improvement before stopping.
        max_bin : int, default=256
            Maximum number of bins for feature discretization.
        max_cat : int, default=1000
            Maximum unique categories before a feature is treated as numerical.
        interaction_constraints : list of list of int, optional
            Groups of feature indices allowed to interact.
        """
        self.treatment_objective = treatment_objective
        self.outcome_objective = outcome_objective
        self.stage1_budget = stage1_budget
        self.stage2_budget = stage2_budget
        self.num_threads = num_threads
        self.monotone_constraints = monotone_constraints
        self.force_children_to_bound_parent = force_children_to_bound_parent
        self.missing = missing
        self.allow_missing_splits = allow_missing_splits
        self.create_missing_branch = create_missing_branch
        self.terminate_missing_features = terminate_missing_features
        self.missing_node_treatment = missing_node_treatment
        self.log_iterations = log_iterations
        self.reset = reset
        self.categorical_features = categorical_features
        self.timeout = timeout
        self.iteration_limit = iteration_limit
        self.memory_limit = memory_limit
        self.stopping_rounds = stopping_rounds
        self.max_bin = max_bin
        self.max_cat = max_cat
        self.interaction_constraints = interaction_constraints
        self.seed = seed

        # Placeholder for standardized constraints and categorical features.
        # These are usually calculated in fit(), but we need to pass something to __init__
        # for CrateIVBooster if we follow the PerpetualBooster pattern of recreating it.
        # However, IVBooster's fit currently uses the same instance.

        self.booster = CrateIVBooster(
            treatment_objective=treatment_objective,
            outcome_objective=outcome_objective,
            stage1_budget=stage1_budget,
            stage2_budget=stage2_budget,
            max_bin=max_bin,
            num_threads=num_threads,
            monotone_constraints={},  # Passed during fit in PerpetualBooster
            interaction_constraints=interaction_constraints,
            force_children_to_bound_parent=force_children_to_bound_parent,
            missing=missing,
            allow_missing_splits=allow_missing_splits,
            create_missing_branch=create_missing_branch,
            terminate_missing_features=set(),
            missing_node_treatment=missing_node_treatment,
            log_iterations=log_iterations,
            seed=seed,
            reset=reset,
            categorical_features=set(),
            timeout=timeout,
            iteration_limit=iteration_limit,
            memory_limit=memory_limit,
            stopping_rounds=stopping_rounds,
        )

    def fit(self, X, Z, y, w) -> Self:
        """
        Fit the IV model.

        Parameters
        ----------
        X : array-like
            Covariates (Controls).
        Z : array-like
            Instruments.
        y : array-like
            Outcome variable.
        w : array-like
            Treatment received.
        """
        # Feature conversion
        features_x, flat_x, rows_x, cols_x, _, _ = convert_input_frame(X, "auto", 1000)
        self.feature_names_in_ = features_x

        # Z conversion (features)
        _, flat_z, rows_z, cols_z, _, _ = convert_input_frame(Z, "auto", 1000)

        if rows_x != rows_z:
            raise ValueError("X and Z must have the same number of rows.")

        # Treatment conversion
        w_, _ = convert_input_array(w, self.treatment_objective)

        # Outcome conversion
        y_, _ = convert_input_array(y, self.outcome_objective)

        self.booster.fit(flat_x, rows_x, cols_x, flat_z, rows_z, cols_z, y_, w_)
        return self

    def predict(self, X, w_counterfactual) -> np.ndarray:
        """
        Predict Outcome given X and a counterfactual W.

        Parameters
        ----------
        X : array-like
            Covariates.
        w_counterfactual : array-like
            Treatment value to simulate.

        Returns
        -------
        preds : ndarray
            Predicted Outcome.
        """
        _, flat_x, rows_x, cols_x, _, _ = convert_input_frame(X, "auto", 1000)

        w_cf_, _ = convert_input_array(
            w_counterfactual, self.treatment_objective
        )  # Use same encoding logic as W

        preds = self.booster.predict(flat_x, rows_x, cols_x, w_cf_)
        return preds

    def to_json(self) -> str:
        """Serialize model to JSON string."""
        return self.booster.json_dump()

    @classmethod
    def from_json(cls, json_str: str) -> "IVBooster":
        """Deserialize model from JSON string."""
        obj = cls.__new__(cls)
        obj.booster = CrateIVBooster.from_json(json_str)
        params = obj.booster.get_params()
        obj.treatment_objective = params["treatment_objective"]
        obj.outcome_objective = params["outcome_objective"]
        obj.stage1_budget = params["stage1_budget"]
        obj.stage2_budget = params["stage2_budget"]
        return obj



"""Meta-learner strategies for heterogeneous treatment effect estimation.

Provides S-Learner, T-Learner, X-Learner, and DR-Learner wrappers around
optimized Rust implementations for estimating the Conditional Average Treatment
Effect (CATE) from observational data.

All learners follow a consistent API:
    - ``fit(X, w, y)`` - fit on covariates, binary treatment, and outcome.
    - ``predict(X)`` - return estimated CATE for each row.
    - ``feature_importances_`` - feature importance from the final effect
      model (available after ``fit``).
"""

from typing import Any, Dict, Optional

import numpy as np
from typing_extensions import Self

from perpetual.perpetual import (
    DRLearner as RustDRLearner,
)
from perpetual.perpetual import (
    SLearner as RustSLearner,
)
from perpetual.perpetual import (
    TLearner as RustTLearner,
)
from perpetual.perpetual import (
    XLearner as RustXLearner,
)


def _validate_binary_treatment(w):
    """Validate that *w* is a binary {0, 1} array."""
    w_arr = np.asarray(w)
    if not np.all(np.isin(w_arr, [0, 1])):
        raise ValueError("Treatment 'w' must be binary (0 or 1).")
    return w_arr.astype(np.float64)


def _prepare_params(budget: float, kwargs: Dict[str, Any]) -> Dict[str, Any]:
    """Prepare parameters for the Rust backend with defaults matching PerpetualBooster."""
    params = {
        "budget": budget,
        "max_bin": 255,
        "num_threads": None,
        "monotone_constraints": {},
        "interaction_constraints": None,
        "force_children_to_bound_parent": False,
        "missing": np.nan,
        "allow_missing_splits": True,
        "create_missing_branch": False,
        "terminate_missing_features": set(),
        "missing_node_treatment": "AssignToParent",
        "log_iterations": 0,
        "seed": 42,
        "reset": None,
        "categorical_features": None,
        "timeout": None,
        "iteration_limit": None,
        "memory_limit": None,
        "stopping_rounds": None,
    }
    params.update(kwargs)
    return params


# ---------------------------------------------------------------------------
# S-Learner
# ---------------------------------------------------------------------------


class SLearner:
    """S-Learner (Single Learner) for Heterogeneous Treatment Effect (HTE) estimation.

    Uses a single model to estimate the outcome:
    ``Y ~ M(X, W)``

    The CATE is obtained by contrasting predictions under treatment and
    control: ``CATE = M(X, 1) - M(X, 0)``.
    """

    def __init__(self, budget: float = 0.5, **kwargs):
        """Create an S-Learner.

        Parameters
        ----------
        budget : float, default=0.5
            Fitting budget forwarded to the Rust backend.
        **kwargs
            Additional keyword arguments forwarded to `PerpetualBooster`.
        """
        self.budget = budget
        self._params = _prepare_params(budget, kwargs)
        self.learner = RustSLearner(**self._params)
        self._n_features: Optional[int] = None

    def fit(self, X, w, y) -> Self:
        """Fit the single model on covariates augmented with treatment.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Covariate matrix.
        w : array-like of shape (n_samples,)
            Binary treatment indicator (0 or 1).
        y : array-like of shape (n_samples,)
            Observed outcome.

        Returns
        -------
        self
            Fitted estimator.
        """
        w_arr = _validate_binary_treatment(w)
        x_arr = np.asarray(X, dtype=float, order="F")
        y_arr = np.asarray(y, dtype=float)

        self._n_features = x_arr.shape[1]
        self.learner.fit(x_arr.ravel(), x_arr.shape[0], x_arr.shape[1], w_arr, y_arr)
        return self

    @property
    def feature_importances_(self) -> np.ndarray:
        """Feature importance of the single model (excluding treatment feature)."""
        importances_map = self.learner.calculate_feature_importance("Gain", True)
        if self._n_features is None:
            return np.zeros(0)
        # Return only the FIRST n_features (the covariates).
        # We assume treatment was appended at index n_features.
        return np.array([importances_map.get(i, 0.0) for i in range(self._n_features)])

    def predict(self, X) -> np.ndarray:
        """Estimate the CATE as ``M(X, 1) - M(X, 0)``.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Covariate matrix.

        Returns
        -------
        ndarray of shape (n_samples,)
            Estimated treatment effect for each sample.
        """
        x_arr = np.asarray(X, dtype=float, order="F")
        return self.learner.predict(x_arr.ravel(), x_arr.shape[0], x_arr.shape[1])


# ---------------------------------------------------------------------------
# T-Learner
# ---------------------------------------------------------------------------


class TLearner:
    """T-Learner (Two Learners) for Heterogeneous Treatment Effect (HTE) estimation.

    Uses two separate models::

        M0(X) ~ Y[W=0]
        M1(X) ~ Y[W=1]

    The CATE is ``M1(X) - M0(X)``.
    """

    def __init__(self, budget: float = 0.5, **kwargs):
        """Create a T-Learner.

        Parameters
        ----------
        budget : float, default=0.5
            Fitting budget forwarded to the Rust backend.
        **kwargs
            Additional keyword arguments forwarded to `PerpetualBooster`.
        """
        self.budget = budget
        self._params = _prepare_params(budget, kwargs)
        self.learner = RustTLearner(**self._params)
        self._n_features: Optional[int] = None

    def fit(self, X, w, y) -> Self:
        x_arr = np.asarray(X, dtype=float, order="F")
        self._n_features = x_arr.shape[1]
        self.learner.fit(
            x_arr.ravel(),
            x_arr.shape[0],
            x_arr.shape[1],
            _validate_binary_treatment(w),
            np.asarray(y, dtype=float),
        )
        return self

    @property
    def feature_importances_(self) -> np.ndarray:
        """Aggregated feature importance from mu0 and mu1."""
        importances_map = self.learner.calculate_feature_importance("Gain", True)
        n = self._n_features if self._n_features is not None else 0
        return np.array([importances_map.get(i, 0.0) for i in range(n)])

    def predict(self, X) -> np.ndarray:
        x_arr = np.asarray(X, dtype=float, order="F")
        return self.learner.predict(x_arr.ravel(), x_arr.shape[0], x_arr.shape[1])


# ---------------------------------------------------------------------------
# X-Learner
# ---------------------------------------------------------------------------


class XLearner:
    """X-Learner for HTE estimation (typically better for imbalanced treatment groups)."""

    def __init__(
        self,
        budget: float = 0.5,
        propensity_budget: Optional[float] = None,
        **kwargs,
    ):
        self.budget = budget
        self._params = _prepare_params(budget, kwargs)
        self._params["propensity_budget"] = propensity_budget
        self.learner = RustXLearner(**self._params)
        self._n_features: Optional[int] = None

    def fit(self, X, w, y) -> Self:
        x_arr = np.asarray(X, dtype=float, order="F")
        self._n_features = x_arr.shape[1]
        self.learner.fit(
            x_arr.ravel(),
            x_arr.shape[0],
            x_arr.shape[1],
            _validate_binary_treatment(w),
            np.asarray(y, dtype=float),
        )
        return self

    @property
    def feature_importances_(self) -> np.ndarray:
        """Aggregated feature importance from the second-stage effect models (tau0, tau1)."""
        importances_map = self.learner.calculate_feature_importance("Gain", True)
        n = self._n_features if self._n_features is not None else 0
        return np.array([importances_map.get(i, 0.0) for i in range(n)])

    def predict(self, X) -> np.ndarray:
        x_arr = np.asarray(X, dtype=float, order="F")
        return self.learner.predict(x_arr.ravel(), x_arr.shape[0], x_arr.shape[1])


# ---------------------------------------------------------------------------
# DR-Learner  (Doubly Robust / AIPW)
# ---------------------------------------------------------------------------


class DRLearner:
    """Doubly Robust (DR) Learner for heterogeneous treatment effect estimation."""

    def __init__(
        self,
        budget: float = 0.5,
        propensity_budget: Optional[float] = None,
        **kwargs,
    ):
        self.budget = budget
        self._params = _prepare_params(budget, kwargs)
        self._params["propensity_budget"] = propensity_budget
        self.learner = RustDRLearner(**self._params)
        self._n_features: Optional[int] = None

    def fit(self, X, w, y) -> Self:
        x_arr = np.asarray(X, dtype=float, order="F")
        self._n_features = x_arr.shape[1]
        self.learner.fit(
            x_arr.ravel(),
            x_arr.shape[0],
            x_arr.shape[1],
            _validate_binary_treatment(w),
            np.asarray(y, dtype=float),
        )
        return self

    @property
    def feature_importances_(self) -> np.ndarray:
        """Feature importance from the outcome model fitted on pseudo-outcomes."""
        importances_map = self.learner.calculate_feature_importance("Gain", True)
        n = self._n_features if self._n_features is not None else 0
        return np.array([importances_map.get(i, 0.0) for i in range(n)])

    def predict(self, X) -> np.ndarray:
        x_arr = np.asarray(X, dtype=float, order="F")
        return self.learner.predict(x_arr.ravel(), x_arr.shape[0], x_arr.shape[1])



import json
import os
import tempfile

import numpy as np
from perpetual import PerpetualBooster


def test_exports_binary():
    X = np.random.randn(50, 2)
    y = np.random.randint(0, 2, 50).astype(float)
    model = PerpetualBooster(objective="LogLoss", budget=0.1)
    model.fit(X, y)

    with tempfile.TemporaryDirectory() as tmp_dir:
        # XGBoost
        xgb_path = os.path.join(tmp_dir, "model.xgb.json")
        model.save_as_xgboost(xgb_path)
        assert os.path.exists(xgb_path)
        with open(xgb_path, "r") as f:
            d = json.load(f)
            assert d["learner"]["objective"]["name"] == "binary:logistic"

        # ONNX
        onnx_path = os.path.join(tmp_dir, "model.onnx")
        model.save_as_onnx(onnx_path)
        assert os.path.exists(onnx_path)


def test_exports_regression():
    X = np.random.randn(50, 2)
    y = np.random.randn(50)
    model = PerpetualBooster(objective="SquaredLoss", budget=0.1)
    model.fit(X, y)

    with tempfile.TemporaryDirectory() as tmp_dir:
        xgb_path = os.path.join(tmp_dir, "model.xgb.json")
        model.save_as_xgboost(xgb_path)
        assert os.path.exists(xgb_path)

        onnx_path = os.path.join(tmp_dir, "model.onnx")
        model.save_as_onnx(onnx_path)
        assert os.path.exists(onnx_path)


def test_exports_multiclass():
    X = np.random.randn(60, 2)
    # Ensure exactly 3 classes
    y = np.array([0, 1, 2] * 20).astype(float)
    model = PerpetualBooster(objective="LogLoss", budget=0.1)
    model.fit(X, y)

    with tempfile.TemporaryDirectory() as tmp_dir:
        xgb_path = os.path.join(tmp_dir, "model.xgb.json")
        model.save_as_xgboost(xgb_path)
        assert os.path.exists(xgb_path)
        with open(xgb_path, "r") as f:
            d = json.load(f)
            assert d["learner"]["objective"]["name"] == "multi:softprob"

        onnx_path = os.path.join(tmp_dir, "model.onnx")
        model.save_as_onnx(onnx_path)
        assert os.path.exists(onnx_path)



import unittest

import numpy as np
from perpetual.iv import IVBooster


class TestIVBooster(unittest.TestCase):
    def setUp(self):
        np.random.seed(42)
        n = 100
        U = np.random.randn(n)  # Unobserved confounder
        z = np.random.randn(n)  # Instrument
        w = 0.8 * z + 0.5 * U + np.random.randn(n) * 0.1
        y = 2.0 * w + 1.2 * U + np.random.randn(n) * 0.1

        self.X = np.random.randn(n, 2)
        self.Z = z.reshape(-1, 1)
        self.y = y
        self.w = w

    def test_initialization(self):
        model = IVBooster(stage1_budget=0.2, stage2_budget=0.3)
        self.assertEqual(model.stage1_budget, 0.2)
        self.assertEqual(model.stage2_budget, 0.3)

    def test_fit_predict(self):
        model = IVBooster(stage1_budget=0.1, stage2_budget=0.1)
        model.fit(self.X, self.Z, self.y, self.w)

        # Standard prediction
        preds = model.predict(self.X, w_counterfactual=self.w)
        self.assertEqual(len(preds), 100)

        # Counterfactual prediction
        preds_cf = model.predict(self.X, w_counterfactual=np.ones(100))
        self.assertEqual(len(preds_cf), 100)

    def test_serialization(self):
        model = IVBooster(stage1_budget=0.1, stage2_budget=0.1)
        model.fit(self.X, self.Z, self.y, self.w)

        json_str = model.to_json()
        self.assertIsInstance(json_str, str)
        self.assertIn("stage1", json_str)
        self.assertIn("stage2", json_str)

        new_model = IVBooster.from_json(json_str)
        new_model = IVBooster.from_json(json_str)
        preds_old = model.predict(self.X, w_counterfactual=self.w)
        preds_new = new_model.predict(self.X, w_counterfactual=self.w)
        np.testing.assert_allclose(preds_old, preds_new, atol=1e-5)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pandas as pd
import pytest
from perpetual import PerpetualBooster
from perpetual.dml import DMLEstimator
from perpetual.fairness import FairClassifier
from perpetual.policy import PolicyLearner
from perpetual.risk import PerpetualRiskEngine


def test_dml_estimator():
    n = 100
    X = np.random.randn(n, 3)
    w = np.random.binomial(1, 0.5, n).astype(float)
    y = X[:, 0] * w + 0.5 * X[:, 1] + np.random.randn(n) * 0.1

    dml = DMLEstimator(budget=0.1, n_folds=2)
    dml.fit(X, w, y)
    preds = dml.predict(X)
    assert preds.shape == (n,)
    assert dml.feature_importances_ is not None


def test_fair_classifier():
    n = 100
    X = np.random.randn(n, 3)
    S = np.random.binomial(1, 0.5, n).astype(float)
    X = np.column_stack([X, S])
    y = (X[:, 0] + 0.5 * S > 0).astype(float)

    clf = FairClassifier(
        sensitive_feature=3, fairness_type="demographic_parity", lam=0.5, budget=0.1
    )
    clf.fit(X, y)

    preds = clf.predict(X)
    probs = clf.predict_proba(X)
    scores = clf.decision_function(X)
    contribs = clf.predict_contributions(X)

    assert preds.shape == (n,)
    assert probs.shape == (n, 2)
    assert scores.shape == (n,)
    assert contribs.shape == (n, 5)

    clf_eo = FairClassifier(
        sensitive_feature=3, fairness_type="equalized_odds", budget=0.1
    )
    clf_eo.fit(X, y)
    assert clf_eo.predict(X).shape == (n,)


def test_policy_learner():
    n = 100
    X = np.random.randn(n, 3)
    w = np.random.binomial(1, 0.5, n)
    y = (X[:, 0] * (w - 0.5) + np.random.randn(n) * 0.1).astype(float)

    pl = PolicyLearner(budget=0.1, mode="ipw")
    pl.fit(X, w, y)
    preds = pl.predict(X)
    probs = pl.predict_proba(X)
    scores = pl.decision_function(X)

    assert preds.shape == (n,)
    assert probs.shape == (n,)
    assert scores.shape == (n,)

    pl_aipw = PolicyLearner(budget=0.1, mode="aipw")
    pl_aipw.fit(X, w, y)
    assert pl_aipw.predict(X).shape == (n,)

    p = np.full(n, 0.5)
    mu1 = np.zeros(n)
    mu0 = np.zeros(n)
    pl_aipw.fit(X, w, y, propensity=p, mu_hat_1=mu1, mu_hat_0=mu0)
    assert pl_aipw.predict(X).shape == (n,)


def test_risk_metrics():
    n = 20
    X = np.random.randn(n, 3)
    y = np.random.randint(0, 2, n).astype(float)

    model = PerpetualBooster(budget=0.1, objective="LogLoss")
    model.fit(X, y)

    engine = PerpetualRiskEngine(model)
    reasons = engine.generate_reason_codes(
        X, threshold=0.5, n_codes=2, rejection_direction="lower"
    )
    assert len(reasons) == n

    model_reg = PerpetualBooster(budget=0.1, objective="SquaredLoss")
    model_reg.fit(X, np.random.randn(n))
    engine_reg = PerpetualRiskEngine(model_reg)
    reasons_reg = engine_reg.generate_reason_codes(
        X, threshold=0.0, n_codes=2, rejection_direction="higher"
    )
    assert len(reasons_reg) == n

    with pytest.raises(ValueError, match="rejection_direction must be"):
        engine_reg.generate_reason_codes(
            X, threshold=0.0, rejection_direction="invalid"
        )


def test_booster_edge_cases():
    X = np.random.randn(100, 2)
    y = np.random.randn(100)

    model = PerpetualBooster(budget=0.1, objective="SquaredLoss", save_node_stats=True)
    model.fit(X, y)

    model.calibrate(X, y, alpha=0.1, method="MinMax")
    model.calibrate(X, y, alpha=[0.1, 0.2], method="GRP")
    model.calibrate(X, y, alpha=0.1, method="WeightVariance")

    X_cal = np.random.randn(50, 2)
    y_cal = np.random.randn(50)
    model.calibrate_conformal(X, y, X_cal, y_cal, alpha=0.1)

    group_invalid = np.array([10, 10])
    with pytest.raises(ValueError, match="Sum of group sizes"):
        model.fit(X, y, group=group_invalid)

    group_ids = np.arange(100)
    with pytest.raises(ValueError, match="Query IDs instead of Group Sizes"):
        model.fit(X, y, group=group_ids)


def test_booster_additional_edge_cases():
    X = np.random.randn(20, 2)
    y = np.random.randn(20)

    model = PerpetualBooster(
        budget=0.1,
        objective="SquaredLoss",
        create_missing_branch=True,
        missing_node_treatment="AssignToParent",
    )
    model.fit(X, y)

    model.prune(X, y)

    with pytest.warns(
        UserWarning, match="predict_proba not implemented for regression"
    ):
        probs = model.predict_proba(X, calibrated=False)
        assert probs.shape == (20, 1)


def test_booster_init_edge_cases():
    with pytest.raises(ValueError, match="fairness_type must be one of"):
        FairClassifier(sensitive_feature=0, fairness_type="invalid")
    with pytest.raises(ValueError, match="n_folds must be >= 2"):
        DMLEstimator(n_folds=1)
    with pytest.raises(ValueError, match="mode must be 'ipw' or 'aipw'"):
        PolicyLearner(mode="invalid")


def test_polars_path_simulation():
    try:
        import polars as pl

        df = pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
        y = np.array([0, 1, 0])
        model = PerpetualBooster(budget=0.1, objective="LogLoss")
        model.fit(df, y)
        preds = model.predict(df)
        assert len(preds) == 3

        X_cal = pl.DataFrame({"a": [1], "b": [4]})
        y_cal = np.array([0])
        model.calibrate_conformal(df, y, X_cal, y_cal, alpha=0.1)

        intervals = model.predict_intervals(df)
        assert len(intervals) >= 0

        sets = model.predict_sets(df)
        assert len(sets) >= 0
    except ImportError:
        pytest.skip("Polars not installed")


def test_utils_conversions():
    from perpetual.utils import (
        convert_input_array,
        convert_input_frame,
        type_df,
        type_series,
    )

    with pytest.raises(ValueError, match="Object type .* is not supported"):
        convert_input_frame("invalid", None, 1000)

    assert type_series(123) == ""
    assert type_df(123) == ""

    # Use numpy array instead of list to avoid AttributeError in fallback path
    arr, cls = convert_input_array(
        np.array([0, 1, 0]), objective="LogLoss", is_target=True
    )
    assert len(cls) == 2

    X = pd.DataFrame({"a": [1, 2, 3], "b": ["cat1", "cat2", "cat1"]})
    # This hits the pandas path in convert_input_frame
    features, flat, rows, cols, cat_idx, cat_map = convert_input_frame(
        X, categorical_features=["b"], max_cat=10
    )
    assert "b" in features


def test_fairness_sigmoid():
    from perpetual.fairness import stable_sigmoid

    assert stable_sigmoid(0) == 0.5
    assert stable_sigmoid(100) > 0.99
    assert stable_sigmoid(-100) < 0.01



"""Internal utilities for data conversion and input validation."""

import logging
from typing import Dict, Iterable, List, Optional, Tuple

import numpy as np

logger = logging.getLogger(__name__)

MIN_AVG_CATEGORY_SUPPORT = 64.0
LOW_CARDINALITY_CATEGORY_SUPPORT = 48.0
LOW_CARDINALITY_CATEGORY_LIMIT = 16
MIN_ROWS_FOR_CATEGORY_SUPPORT_CHECK = 128


def _is_polars_string_like_dtype(dtype, pl_module) -> bool:
    base_type = dtype.base_type() if hasattr(dtype, "base_type") else dtype
    if base_type in {pl_module.Categorical, pl_module.String}:
        return True

    enum_dtype = getattr(pl_module, "Enum", None)
    return enum_dtype is not None and base_type == enum_dtype


CLASSIFICATION_OBJECTIVES = {
    "LogLoss",
    "BrierLoss",
    "HingeLoss",
    "CrossEntropyLoss",
    "CrossEntropyLambdaLoss",
}


def categorical_support_threshold(
    n_non_missing_categories: int, n_non_missing_rows: int
) -> float:
    if n_non_missing_rows < MIN_ROWS_FOR_CATEGORY_SUPPORT_CHECK:
        return 0.0
    if n_non_missing_categories <= LOW_CARDINALITY_CATEGORY_LIMIT:
        return LOW_CARDINALITY_CATEGORY_SUPPORT
    return MIN_AVG_CATEGORY_SUPPORT


def type_df(df):
    """Return a string tag for the DataFrame library (``'pandas_df'``, ``'polars_df'``, ``'numpy'``, or ``''``)."""
    library_name = type(df).__module__.split(".")[0]
    if type(df).__name__ == "DataFrame":
        if library_name == "pandas":
            return "pandas_df"
        elif library_name == "polars":
            return "polars_df"
    elif library_name == "numpy":
        return "numpy"
    else:
        return ""


def type_series(y):
    """Return a string tag for the Series library (``'pandas_series'``, ``'polars_series'``, ``'numpy'``, or ``''``)."""
    library_name = type(y).__module__.split(".")[0]
    if type(y).__name__ == "Series":
        if library_name == "pandas":
            return "pandas_series"
        elif library_name == "polars":
            return "polars_series"
    elif library_name == "numpy":
        return "numpy"
    else:
        return ""


def convert_input_array(
    x, objective, is_target=False, is_int=False, is_classification=None
) -> np.ndarray:
    """Convert an array-like input to a flat ``float64`` (or ``uint64``) NumPy array.

    Returns
    -------
    tuple of (np.ndarray, list)
        The converted array and the detected class labels (empty list if not classification).
    """
    classes_ = []

    if type(x).__module__.split(".")[0] == "numpy":
        if len(x.shape) == 2:
            classes_, x_, *_ = convert_input_frame(x, None, 1000)
        else:
            x_ = x
    elif type_series(x) == "pandas_series":
        x_ = x.to_numpy()
    elif type_series(x) == "polars_series":
        x_ = x.to_numpy(allow_copy=False)
    elif type_df(x) == "polars_df" or type_df(x) == "pandas_df":
        classes_, x_, *_ = convert_input_frame(x, None, 1000)
    else:
        x_ = x.to_numpy()

    # Determine if we should perform class detection
    do_class_detection = False
    if is_target and len(x_.shape) == 1:
        if is_classification is True:
            do_class_detection = True
        elif is_classification is False:
            do_class_detection = False
        else:
            # Fallback to heuristic if not explicitly set
            do_class_detection = objective in CLASSIFICATION_OBJECTIVES

    if do_class_detection:
        classes_, x_index = np.unique(x_, return_inverse=True)

        if len(classes_) > 1:
            if len(classes_) > 2:
                x_ = np.eye(len(classes_))[x_index]
            else:
                x_ = x_index.astype("float64")

    if is_int and not np.issubdtype(x_.dtype, "uint64"):
        x_ = x_.astype(dtype="uint64", copy=False)

    if not is_int and not np.issubdtype(x_.dtype, "float64"):
        x_ = x_.astype(dtype="float64", copy=False)

    if len(x_.shape) == 2:
        x_ = x_.ravel(order="F")

    return x_, classes_


def convert_input_frame(
    X,
    categorical_features,
    max_cat,
) -> Tuple[List[str], np.ndarray, int, int, Optional[Iterable[int]], Optional[Dict]]:
    """Convert data to format needed by booster.

    Returns:
        Tuple[List[str], np.ndarray, int, int, Optional[Iterable[int]], Optional[Dict]]: Return column names, the flat data, number of rows, the number of columns, cat_index, cat_mapping
    """
    categorical_features_ = None
    if type_df(X) == "pandas_df":
        X_ = X.to_numpy()
        features_ = X.columns.to_list()
        if categorical_features == "auto":
            categorical_columns = X.select_dtypes(
                include=["category", "object", "string"]
            ).columns.tolist()
            categorical_features_ = [
                features_.index(c) for c in categorical_columns
            ] or None
    elif type_df(X) == "numpy":
        X_ = X
        features_ = list(map(str, range(X_.shape[1])))
    else:
        raise ValueError(f"Object type {type(X)} is not supported.")

    if (
        categorical_features
        and all(isinstance(s, int) for s in categorical_features)
        and isinstance(categorical_features, list)
    ):
        categorical_features_ = categorical_features
    elif (
        categorical_features
        and all(isinstance(s, str) for s in categorical_features)
        and isinstance(categorical_features, list)
    ):
        categorical_features_ = [features_.index(c) for c in categorical_features]

    cat_mapping = {}  # key: feature_name, value: ordered category names
    cat_to_num = []
    if categorical_features_:
        for i in categorical_features_:
            categories, inversed = np.unique(X_[:, i].astype(str), return_inverse=True)

            categories = list(categories)
            if "nan" in categories:
                categories.remove("nan")
            n_non_missing_categories = len(categories)
            n_non_missing_rows = np.count_nonzero(
                ~np.isnan(X_[:, i].astype(dtype="float64", copy=False))
                if np.issubdtype(X_[:, i].dtype, np.number)
                else X_[:, i].astype(str) != "nan"
            )
            avg_category_support = (
                n_non_missing_rows / n_non_missing_categories
                if n_non_missing_categories > 0
                else float("inf")
            )
            min_support_threshold = categorical_support_threshold(
                n_non_missing_categories, n_non_missing_rows
            )
            categories.insert(0, "nan")

            inversed = inversed + 1.0

            if n_non_missing_categories > max_cat:
                cat_to_num.append(i)
                logger.warning(
                    f"Feature {features_[i]} will be treated as numerical since the number of non-missing categories ({n_non_missing_categories}) exceeds max_cat ({max_cat}) threshold."
                )
            elif avg_category_support < min_support_threshold:
                cat_to_num.append(i)
                logger.warning(
                    f"Feature {features_[i]} will be treated as numerical since its average support per non-missing category ({avg_category_support:.1f}) is below the native categorical stability threshold ({min_support_threshold:.1f})."
                )

            feature_name = features_[i]
            cat_mapping[feature_name] = categories
            ind_nan = len(categories)
            inversed[inversed == ind_nan] = np.nan
            X_[:, i] = inversed

        categorical_features_ = [
            x for x in categorical_features_ if x not in cat_to_num
        ]

        logger.info(f"Categorical features: {categorical_features_}")
        logger.info(f"Mapping of categories: {cat_mapping}")

    if not np.issubdtype(X_.dtype, "float64"):
        X_ = X_.astype(dtype="float64", copy=False)
    flat_data = X_.ravel(order="F")
    rows, cols = X_.shape

    if isinstance(categorical_features_, list):
        categorical_features_ = set(categorical_features_)

    return features_, flat_data, rows, cols, categorical_features_, cat_mapping


def convert_input_frame_columnar(
    X, categorical_features, max_cat
) -> Tuple[
    List[str],
    List[np.ndarray],
    List[Optional[np.ndarray]],
    int,
    int,
    Optional[set],
    dict,
]:
    """Convert Polars DataFrame to columnar format for zero-copy transfer.

    Returns list of column arrays and list of validity masks.
    """
    import polars as pl

    features_ = list(X.columns)
    rows, cols = X.shape

    # Determine categorical features
    categorical_features_ = None
    if categorical_features == "auto":
        categorical_columns = [
            name
            for name, dtype in X.schema.items()
            if _is_polars_string_like_dtype(dtype, pl)
        ]
        categorical_features_ = [
            features_.index(c) for c in categorical_columns
        ] or None
    elif (
        categorical_features
        and all(isinstance(s, int) for s in categorical_features)
        and isinstance(categorical_features, list)
    ):
        categorical_features_ = categorical_features
    elif (
        categorical_features
        and all(isinstance(s, str) for s in categorical_features)
        and isinstance(categorical_features, list)
    ):
        categorical_features_ = [features_.index(c) for c in categorical_features]

    cat_mapping = {}
    cat_to_num = []
    categorical_set = set(categorical_features_) if categorical_features_ else set()

    # Convert each column to numpy array
    columns = []
    masks = []
    for i, col_name in enumerate(features_):
        if i in categorical_set:
            # For categorical columns, cast to String first to bypass
            # Polars' global string cache which can make the Arrow
            # DictionaryArray dictionary unreliable (it may contain
            # categories from ALL categorical columns, not just this one).
            str_series = X[col_name].cast(pl.String)
            str_np = str_series.to_numpy(allow_copy=True)
            null_mask = str_series.is_null().to_numpy()

            # Get sorted unique non-null categories (consistent with Pandas np.unique path)
            valid_strs = str_np[~null_mask]
            if len(valid_strs) > 0:
                cats = sorted(set(valid_strs))
                cats = [c for c in cats if c != "nan"]
            else:
                cats = []
            n_non_missing_categories = len(cats)
            n_non_missing_rows = int((~null_mask).sum())
            avg_category_support = (
                n_non_missing_rows / n_non_missing_categories
                if n_non_missing_categories > 0
                else float("inf")
            )
            min_support_threshold = categorical_support_threshold(
                n_non_missing_categories, n_non_missing_rows
            )

            cats.insert(0, "nan")

            if n_non_missing_categories > max_cat:
                cat_to_num.append(i)
                logger.warning(
                    f"Feature {col_name} will be treated as numerical since the number of non-missing categories ({n_non_missing_categories}) exceeds max_cat ({max_cat}) threshold."
                )
            elif avg_category_support < min_support_threshold:
                cat_to_num.append(i)
                logger.warning(
                    f"Feature {col_name} will be treated as numerical since its average support per non-missing category ({avg_category_support:.1f}) is below the native categorical stability threshold ({min_support_threshold:.1f})."
                )

            cat_mapping[col_name] = cats

            # Encode: use searchsorted on sorted cats (without "nan" at 0) for vectorized mapping
            cats_no_nan = cats[1:]  # remove "nan" prefix
            cats_arr = np.array(cats_no_nan)
            out_values = np.full(rows, np.nan, dtype=np.float64)
            valid_indices = ~null_mask
            if valid_indices.any():
                positions = np.searchsorted(cats_arr, valid_strs)
                encoded = positions.astype(np.float64) + 1.0
                exact_match = np.zeros(len(valid_strs), dtype=bool)
                in_range = positions < len(cats_arr)
                if in_range.any():
                    exact_match[in_range] = (
                        cats_arr[positions[in_range]] == valid_strs[in_range]
                    )
                encoded[~exact_match] = np.nan
                out_values[valid_indices] = encoded

            columns.append(out_values)
            masks.append(None)  # Categorical encoding handles NaNs
        else:
            series = X[col_name].cast(pl.Float64)
            columns.append(
                series.to_numpy(allow_copy=True).astype(np.float64, copy=False)
            )
            masks.append(None)

    if categorical_features_:
        categorical_features_ = [
            x for x in categorical_features_ if x not in cat_to_num
        ]
        logger.info(f"Categorical features: {categorical_features_}")
        logger.info(f"Mapping of categories: {cat_mapping}")

    if isinstance(categorical_features_, list):
        categorical_features_ = set(categorical_features_)

    return features_, columns, masks, rows, cols, categorical_features_, cat_mapping


def transform_input_frame(X, cat_mapping) -> Tuple[List[str], np.ndarray, int, int]:
    """Convert data to format needed by booster.

    Returns:
        Tuple[List[str], np.ndarray, int, int]: Return column names, the flat data, number of rows, the number of columns
    """
    if type_df(X) == "pandas_df":
        X_ = X.to_numpy()
        features_ = X.columns.to_list()
    elif type_df(X) == "numpy":
        X_ = X
        features_ = list(map(str, range(X_.shape[1])))
    else:
        raise ValueError(f"Object type {type(X)} is not supported.")

    if cat_mapping:
        for feature_name, categories in cat_mapping.items():
            feature_index = features_.index(feature_name)
            cats = categories.copy()
            cats.remove("nan")
            values = X_[:, feature_index].astype(str)
            positions = np.searchsorted(cats, values)
            x_enc = positions.astype(np.float64) + 1.0
            exact_match = np.zeros(len(values), dtype=bool)
            in_range = positions < len(cats)
            if in_range.any():
                cat_array = np.asarray(cats, dtype=object)
                exact_match[in_range] = (
                    cat_array[positions[in_range]] == values[in_range]
                )
            x_enc[~exact_match] = np.nan
            X_[:, feature_index] = x_enc

    if not np.issubdtype(X_.dtype, "float64"):
        X_ = X_.astype(dtype="float64", copy=False)
    flat_data = X_.ravel(order="F")
    rows, cols = X_.shape

    return features_, flat_data, rows, cols


def transform_input_frame_columnar(
    X, cat_mapping
) -> Tuple[List[str], List[np.ndarray], List[Optional[np.ndarray]], int, int]:
    """Convert Polars DataFrame to columnar format for zero-copy prediction.

    Returns list of column arrays and masks instead of flattened data, avoiding copies.
    """
    features_ = list(X.columns)
    rows, cols = X.shape

    columns = []
    masks = []
    import polars as pl

    for i, col_name in enumerate(features_):
        if cat_mapping and col_name in cat_mapping:
            # For categorical columns, encode using the existing cat_mapping.
            # Cast to String first to bypass Polars' global string cache
            # which can make Arrow DictionaryArray dictionaries unreliable.
            categories = cat_mapping[col_name]
            cats = categories.copy()
            cats.remove("nan")

            str_series = X[col_name].cast(pl.String)
            str_np = str_series.to_numpy(allow_copy=True)
            null_mask = str_series.is_null().to_numpy()

            x_enc = np.full(rows, np.nan, dtype=np.float64)
            valid_indices = ~null_mask
            if valid_indices.any():
                valid_strs = str_np[valid_indices]
                positions = np.searchsorted(cats, valid_strs)
                encoded = positions.astype(np.float64) + 1.0
                exact_match = np.zeros(len(valid_strs), dtype=bool)
                in_range = positions < len(cats)
                if in_range.any():
                    cat_array = np.asarray(cats, dtype=object)
                    exact_match[in_range] = (
                        cat_array[positions[in_range]] == valid_strs[in_range]
                    )
                encoded[~exact_match] = np.nan
                x_enc[valid_indices] = encoded

            columns.append(x_enc)
            masks.append(None)
        else:
            series = X[col_name].cast(pl.Float64)
            columns.append(
                series.to_numpy(allow_copy=True).astype(np.float64, copy=False)
            )
            masks.append(None)

    return features_, columns, masks, rows, cols


CONTRIBUTION_METHODS = {
    "weight": "Weight",
    "Weight": "Weight",
    "average": "Average",
    "Average": "Average",
    "branch-difference": "BranchDifference",
    "branchdifference": "BranchDifference",
    "BranchDifference": "BranchDifference",
    "midpoint-difference": "MidpointDifference",
    "midpointdifference": "MidpointDifference",
    "MidpointDifference": "MidpointDifference",
    "mode-difference": "ModeDifference",
    "modedifference": "ModeDifference",
    "ModeDifference": "ModeDifference",
    "ProbabilityChange": "ProbabilityChange",
    "probabilitychange": "ProbabilityChange",
    "probability-change": "ProbabilityChange",
}



import os
from typing import Tuple

import numpy as np
import pandas as pd
import pytest
from perpetual.booster import PerpetualBooster
from perpetual.sklearn import PerpetualClassifier, PerpetualRanker, PerpetualRegressor
from sklearn.datasets import make_classification, make_regression
from sklearn.model_selection import train_test_split


@pytest.fixture
def X_y() -> Tuple[pd.DataFrame, pd.Series]:
    df = pd.read_csv(
        os.path.join(os.path.dirname(__file__), "../../../resources", "titanic.csv")
    )
    X = df.select_dtypes("number").drop(columns="survived").reset_index(drop=True)
    y = df["survived"]
    return X, y


@pytest.fixture
def X_y_g() -> Tuple[pd.DataFrame, pd.Series, pd.Series]:
    df = pd.read_csv(
        os.path.join(os.path.dirname(__file__), "../../../resources", "goodreads.csv")
    )

    df["group"] = df["year"].astype(str) + "_" + df["category"]

    df = df.sort_values("group")

    composite_groups = df["group"]
    group_ids, _unique_groups = pd.factorize(composite_groups)

    group_lengths = pd.Series(group_ids).value_counts().sort_index()

    feature_cols = [
        "avg_rating",
        "pages",
        "5stars",
        "4stars",
        "3stars",
        "2stars",
        "1stars",
        "ratings",
    ]
    target_col = "rank"

    X = df[feature_cols]

    rank = df[target_col]
    y = rank.max() - rank

    return X, y, group_lengths


def test_sklearn_compat_classification(X_y):
    X, y = X_y
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=42)

    model_sklearn = PerpetualClassifier(objective="LogLoss", iteration_limit=50)
    model_sklearn.fit(X_train, y_train)

    preds_sklearn = model_sklearn.predict(X_test)
    proba_sklearn = model_sklearn.predict_proba(X_test)
    log_odds_sklearn = model_sklearn.predict_log_proba(X_test)

    model_booster = PerpetualBooster(objective="LogLoss", iteration_limit=50)
    model_booster.fit(X_train, y_train)

    preds_booster = model_booster.predict(X_test)
    proba_booster = model_booster.predict_proba(X_test)
    log_odds_booster = model_booster.predict_log_proba(X_test)

    assert np.allclose(preds_sklearn, preds_booster)
    assert np.allclose(proba_sklearn, proba_booster)
    assert np.allclose(log_odds_sklearn, log_odds_booster)


def test_sklearn_compat_regression(X_y):
    X, y = X_y
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=42)

    model_sklearn = PerpetualRegressor(objective="SquaredLoss", iteration_limit=50)
    model_sklearn.fit(X_train, y_train)

    preds_sklearn = model_sklearn.predict(X_test)

    model_booster = PerpetualBooster(objective="SquaredLoss", iteration_limit=50)
    model_booster.fit(X_train, y_train)

    preds_booster = model_booster.predict(X_test)

    assert np.allclose(preds_sklearn, preds_booster)


def test_sklearn_compat_ranking(X_y_g):
    X, y, group_lengths = X_y_g

    model_sklearn = PerpetualRanker(objective="ListNetLoss", iteration_limit=50)
    model_sklearn.fit(X, y, group=group_lengths)

    preds_sklearn = model_sklearn.predict(X)

    model_booster = PerpetualBooster(objective="ListNetLoss", iteration_limit=50)
    model_booster.fit(X, y, group=group_lengths)

    preds_booster = model_booster.predict(X)

    assert np.allclose(preds_sklearn, preds_booster)


def test_perpetual_ranker():
    n = 100
    X = np.random.randn(n, 3)
    y = np.random.randn(n)
    # 5 groups of 20
    group = np.array([20] * 5)

    ranker = PerpetualRanker(budget=0.1)
    ranker.fit(X, y, group=group)
    preds = ranker.predict(X)
    assert preds.shape == (n,)
    assert isinstance(ranker.score(X, y), float)


def test_perpetual_classifier_score():
    X, y = make_classification(n_samples=100, n_features=10, random_state=42)
    clf = PerpetualClassifier(iteration_limit=10)
    clf.fit(X, y)
    score = clf.score(X, y)
    assert isinstance(score, float)
    assert 0 <= score <= 1.0


def test_perpetual_classifier_warning():
    X, y = make_classification(n_samples=100, n_features=10, random_state=42)
    # Using a regression objective for a classifier should trigger a warning
    clf = PerpetualClassifier(objective="SquaredLoss", iteration_limit=10)
    with pytest.warns(
        UserWarning, match="Objective 'SquaredLoss' is typically for regression/ranking"
    ):
        clf.fit(X, y)


def test_perpetual_regressor_score():
    X, y = make_regression(n_samples=100, n_features=10, random_state=42)
    reg = PerpetualRegressor(iteration_limit=10)
    reg.fit(X, y)
    score = reg.score(X, y)
    assert isinstance(score, float)
    # R2 score can be negative, but should be <= 1.0


def test_perpetual_regressor_warning():
    X, y = make_regression(n_samples=100, n_features=10, random_state=42)
    # Using a classification objective for a regressor should trigger a warning
    reg = PerpetualRegressor(objective="LogLoss", iteration_limit=10)
    with pytest.warns(
        UserWarning,
        match="Objective 'LogLoss' may not be suitable for PerpetualRegressor",
    ):
        reg.fit(X, y)


def test_perpetual_ranker_init_warning():
    # Using a classification objective for a ranker should trigger a warning in __init__
    with pytest.warns(
        UserWarning, match="Objective 'LogLoss' may not be suitable for PerpetualRanker"
    ):
        PerpetualRanker(objective="LogLoss")


def test_perpetual_ranker_fit_error():
    X = np.random.randn(100, 10)
    y = np.random.randn(100)
    ranker = PerpetualRanker(objective="ListNetLoss", iteration_limit=10)
    # ListNetLoss requires group parameter
    with pytest.raises(ValueError, match="The 'group' parameter must be provided"):
        ranker.fit(X, y)


def test_perpetual_classifier_custom_objective():
    X, y = make_classification(n_samples=10, n_features=10, random_state=42)
    clf = PerpetualClassifier(objective="CustomStringObjective", iteration_limit=1)
    # We just want to hit the line in sklearn.py __init__
    try:
        clf.fit(X, y)
    except Exception:
        pass


def test_perpetual_regressor_custom_objective():
    X, y = make_regression(n_samples=10, n_features=10, random_state=42)
    reg = PerpetualRegressor(objective="CustomStringObjective", iteration_limit=1)
    try:
        reg.fit(X, y)
    except Exception:
        pass



import os
import re
import sys

import requests


def parse_description(package_dir):
    """Parses DESCRIPTION file for package name, version, and maintainer."""
    desc_path = os.path.join(package_dir, "DESCRIPTION")
    with open(desc_path, "r", encoding="utf-8") as f:
        content = f.read()

    pkg = re.search(r"^Package:\s*(.+)", content, re.MULTILINE).group(1)
    version = re.search(r"^Version:\s*(.+)", content, re.MULTILINE).group(1)

    maintainer_match = re.search(
        r"Maintainer:\s*([^<]+)<([^>]+)>", content, re.MULTILINE
    )
    if not maintainer_match:
        # Try Authors@R
        authors_match = re.search(
            r"Authors@R:\s*(.+)", content, re.MULTILINE | re.DOTALL
        )
        if authors_match:
            # Very basic parsing of Authors@R to find maintainer
            # Looking for role = "cre" or role = c(..., "cre", ...)
            authors = authors_match.group(1)
            # This is a bit fragile but covers standard cases
            cre_match = re.search(
                r'person\("([^"]+)",\s*"([^"]+)",\s*email\s*=\s*"([^"]+)",.*role.*"cre"',
                authors,
            )
            if cre_match:
                name = f"{cre_match.group(1)} {cre_match.group(2)}"
                email = cre_match.group(3)
            else:
                raise ValueError("Could not find Maintainer in DESCRIPTION")
        else:
            raise ValueError("Could not find Maintainer in DESCRIPTION")
    else:
        name = maintainer_match.group(1).strip()
        email = maintainer_match.group(2).strip()

    print(f"Detected Package: {pkg} {version}")
    print(f"Detected Maintainer: {name} <{email}>")

    return pkg, version, name, email


def submit_cran(tarball_path, package_dir):
    pkg, version, name, email = parse_description(package_dir)

    # Read comment
    comment = f"Releasing version {version}"
    comments_path = os.path.join(package_dir, "cran-comments.md")
    if os.path.exists(comments_path):
        with open(comments_path, "r", encoding="utf-8") as f:
            comment = f.read()

    url = "https://xmpalantir.wu.ac.at/cransubmit/index2.php"

    files = {
        "uploaded_file": (
            os.path.basename(tarball_path),
            open(tarball_path, "rb"),
            "application/x-gzip",
        ),
    }

    data = {
        "pkg_id": "",
        "name": name,
        "email": email,
        "comment": comment,
        "upload": "Upload the package",
    }

    print(f"Submitting {tarball_path} to CRAN (Step 1: Upload)...")
    try:
        session = requests.Session()
        response = session.post(url, files=files, data=data)
        response.raise_for_status()

        # Step 2: Check for confirmation page and extract pkg_id
        if "Please review and submit" in response.text:
            print("Step 1 successful. Proceeding to Step 2 (Confirmation)...")

            # Extract pkg_id
            pkg_id_match = re.search(
                r'name="pkg_id" type="hidden" value="([^"]+)"', response.text
            )
            if not pkg_id_match:
                print("Error: Could not find pkg_id in confirmation page.")
                print(response.text)
                sys.exit(1)

            pkg_id = pkg_id_match.group(1)
            print(f"Found pkg_id: {pkg_id}")

            # Prepare data for Step 2
            confirm_data = {
                "pkg_id": pkg_id,
                "name": name,
                "email": email,
                "policy_check": "1",  # Policy check might be needed here or just implicit
                "submit": "Submit package",
            }

            print("Submitting confirmation...")
            response = session.post(url, data=confirm_data)
            response.raise_for_status()

        # Check for final success message
        # Success message contains HTML tags: "The <b>maintainer of this package</b> has been sent an email..."
        # So we check for a unique substring without tags.
        if (
            "The following package has been uploaded:" in response.text
            or "has been sent an email to confirm the submission" in response.text
        ):
            print("Submission successful!")
            print("Check your email for the confirmation link.")
        else:
            print("Submission might have failed. Response content:")
            print(response.text)
            sys.exit(1)

    except Exception as e:
        print(f"Error submitting to CRAN: {e}")
        sys.exit(1)


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: python submit_cran.py <tarball_path>")
        sys.exit(1)

    tarball = sys.argv[1]
    # Assuming script is run from project root and package-r is there
    package_dir = "package-r"

    submit_cran(tarball, package_dir)



"""Tests for zero-copy Polars integration."""

import tracemalloc

import numpy as np
import polars as pl
from perpetual import PerpetualBooster


def test_polars_zero_copy_fit():
    """Verify fitting with Polars DataFrame doesn't copy data."""
    n_rows, n_cols = 1000, 10
    data = {f"col_{i}": np.random.randn(n_rows) for i in range(n_cols)}
    df = pl.DataFrame(data)
    y = np.random.randint(0, 2, n_rows).astype(np.float64)
    data_size_bytes = n_rows * n_cols * 8

    # Ensure everything is pre-allocated
    _ = df.to_numpy()

    tracemalloc.start()
    snapshot_before = tracemalloc.take_snapshot()
    model = PerpetualBooster(objective="LogLoss", budget=0.1, iteration_limit=1)
    model.fit(df, y)
    snapshot_after = tracemalloc.take_snapshot()
    tracemalloc.stop()

    stats = snapshot_after.compare_to(snapshot_before, "lineno")
    for stat in stats[:5]:
        print(stat)
    total_increase = sum(stat.size_diff for stat in stats if stat.size_diff > 0)

    # dataset is ~80KB.
    # If this fails, it might be due to initializations.
    # We allow more buffer for first run initialization.
    assert total_increase < data_size_bytes * 10.0 + 1_000_000


def test_polars_zero_copy_predict():
    """Verify predicting with Polars DataFrame doesn't copy data."""
    n_rows, n_cols = 1000, 10
    data = {f"col_{i}": np.random.randn(n_rows) for i in range(n_cols)}
    df = pl.DataFrame(data)
    y = np.random.randint(0, 2, n_rows).astype(np.float64)

    model = PerpetualBooster(objective="LogLoss", budget=0.1, iteration_limit=1)
    model.fit(df, y)

    tracemalloc.start()
    _ = model.predict(df)
    tracemalloc.take_snapshot()
    _, peak = tracemalloc.get_traced_memory()
    tracemalloc.stop()

    # Peak should not include a full copy of X (80KB)
    assert peak < n_rows * n_cols * 8 * 1.5


def test_polars_fit_predict_correctness():
    """Verify that Polars columnar path produces same results as NumPy."""
    n_rows, n_cols = 200, 5
    data = {f"col_{i}": np.linspace(0, 1, n_rows) + i for i in range(n_cols)}
    df_polars = pl.DataFrame(data)
    arr_numpy = df_polars.to_numpy()
    y = (df_polars["col_0"] > 0.5).cast(pl.Float64).to_numpy()

    # Fit with Polars
    model_polars = PerpetualBooster(objective="LogLoss", budget=0.1, iteration_limit=5)
    model_polars.fit(df_polars, y)
    preds_polars = model_polars.predict(df_polars)

    # Fit with NumPy
    model_numpy = PerpetualBooster(objective="LogLoss", budget=0.1, iteration_limit=5)
    model_numpy.fit(arr_numpy, y)
    preds_numpy = model_numpy.predict(arr_numpy)

    assert model_polars.number_of_trees > 0
    assert np.allclose(preds_polars, preds_numpy, atol=1e-5)


def test_polars_zero_copy_predict_contributions():
    """Verify predicting contributions with Polars."""
    n_rows, n_cols = 100, 5
    data = {f"col_{i}": np.random.randn(n_rows) for i in range(n_cols)}
    df = pl.DataFrame(data)
    y = np.random.randint(0, 2, n_rows).astype(np.float64)

    model = PerpetualBooster(objective="LogLoss", budget=0.1, iteration_limit=1)
    model.fit(df, y)

    contribs = model.predict_contributions(df)
    assert contribs.shape == (n_rows, n_cols + 1)


def test_polars_multi_output():
    """Verify multi-output with Polars."""
    n_rows, n_cols = 100, 5
    data = {f"col_{i}": np.random.randn(n_rows) for i in range(n_cols)}
    df = pl.DataFrame(data)
    y = np.random.randint(0, 3, n_rows)

    model = PerpetualBooster(objective="LogLoss", budget=0.1, iteration_limit=1)
    model.fit(df, y)

    preds = model.predict(df)
    assert len(preds) == n_rows

    proba = model.predict_proba(df)
    assert proba.shape == (n_rows, 3)

    contribs = model.predict_contributions(df)
    # 3 classes * (5 features + 1 bias) = 18 columns
    assert contribs.shape == (n_rows, 3 * (n_cols + 1))



import numpy as np
from perpetual import IVBooster, UpliftBooster


def test_braided_interaction_constraints():
    X = np.random.rand(100, 5)
    Z = np.random.rand(100, 2)
    y = np.random.rand(100)
    w = np.random.rand(100)

    interaction_constraints = [[0, 1], [2, 3, 4]]

    model = IVBooster(
        stage1_budget=0.1,
        stage2_budget=0.1,
        interaction_constraints=interaction_constraints,
    )
    model.fit(X, Z, y, w)

    # Check if constraints are stored in the underlying crate objects
    # We can check via json_dump
    json_str = model.to_json()
    assert "interaction_constraints" in json_str


def test_uplift_interaction_constraints():
    X = np.random.rand(100, 5)
    y = np.random.rand(100)
    # Treatment must be binary 0 or 1
    w = np.random.randint(0, 2, 100)

    interaction_constraints = [[0, 2], [1, 3, 4]]

    model = UpliftBooster(
        outcome_budget=0.1,
        propensity_budget=0.1,
        effect_budget=0.1,
        interaction_constraints=interaction_constraints,
    )
    model.fit(X, w, y)

    json_str = model.to_json()
    assert "interaction_constraints" in json_str


if __name__ == "__main__":
    test_braided_interaction_constraints()
    test_uplift_interaction_constraints()



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from perpetual import PerpetualBooster
from sklearn.datasets import fetch_california_housing
from sklearn.model_selection import train_test_split


def run_calibration_demo():
    # 1. Load data
    print("Loading California Housing data...")
    data = fetch_california_housing()
    X = pd.DataFrame(data.data, columns=data.feature_names)
    y = data.target

    # 2. Split data: 60% train, 20% calibration, 20% test
    X_train_full, X_test, y_train_full, y_test = train_test_split(
        X, y, test_size=0.2, random_state=42
    )
    X_train, X_cal, y_train, y_cal = train_test_split(
        X_train_full, y_train_full, test_size=0.25, random_state=42
    )

    print(
        f"Train size: {len(X_train)}, Calibration size: {len(X_cal)}, Test size: {len(X_test)}"
    )

    # 3. Define calibration methods to test
    methods = ["Conformal", "MinMax", "GRP", "WeightVariance"]
    alpha = 0.1  # 90% coverage target

    results = {}

    for method in methods:
        print(f"\n--- Method: {method} ---")

        # Initialize booster
        model = PerpetualBooster(
            objective="SquaredLoss", budget=1.0, save_node_stats=True
        )

        # Fit the model
        print("Fitting model...")
        model.fit(X_train, y_train)

        # Calibrate
        print("Calibrating...")
        if method == "Conformal":
            model.calibrate_conformal(X_train, y_train, X_cal, y_cal, alpha=[alpha])
        else:
            model.calibrate(X_cal, y_cal, alpha=[alpha], method=method)

        # Predict intervals on test set
        print("Predicting intervals...")
        intervals = model.predict_intervals(X_test)

        # Calculate coverage and average width
        lower = intervals[str(alpha)][:, 0]
        upper = intervals[str(alpha)][:, 1]

        covered = np.mean((y_test >= lower) & (y_test <= upper))
        avg_width = np.mean(upper - lower)

        print(f"Coverage: {covered:.2%}")
        print(f"Average Width: {avg_width:.4f}")

        results[method] = {
            "coverage": covered,
            "width": avg_width,
            "lower": lower,
            "upper": upper,
        }

    # 4. Simple visualization of first 50 test samples
    plt.figure(figsize=(12, 6))
    plt.plot(y_test[:50], "k.", label="Actual Values", alpha=0.6)

    colors = ["r", "g", "b", "orange"]
    for i, method in enumerate(methods):
        plt.plot(
            results[method]["lower"][:50], color=colors[i], linestyle="--", alpha=0.5
        )
        plt.plot(
            results[method]["upper"][:50],
            color=colors[i],
            linestyle="--",
            alpha=0.5,
            label=f"{method} bounds",
        )

    plt.title("Prediction Intervals (90% target) - First 50 Test Samples")
    plt.xlabel("Sample Index")
    plt.ylabel("MedHouseVal")
    plt.legend()
    plt.grid(True, alpha=0.3)
    plt.show()

    print("\nSummary Results:")
    for method, metrics in results.items():
        print(
            f"{method:15}: Coverage={metrics['coverage']:.2%}, Avg Width={metrics['width']:.4f}"
        )


if __name__ == "__main__":
    run_calibration_demo()



"""
Compare PerpetualBooster and LightGBM on wind_onshore.csv time series (stationary, lag features).

uv run python package-python/examples/benchmark_wind_onshore.py
"""

import os
from time import process_time, time

import pandas as pd
from lightgbm import LGBMRegressor
from perpetual import PerpetualBooster
from sklearn.metrics import mean_squared_error


# --- Data Preparation ---
def load_wind_onshore_stationary(csv_path, test_size=0.2, lags=24):
    df = pd.read_csv(csv_path, sep=";")
    df["datetime"] = pd.to_datetime(df["Date/Time CET/CEST"], format="%d.%m.%Y/%H:%M")
    df = df.sort_values("datetime").reset_index(drop=True)
    df["wind_onshore"] = (
        df["Wind Onshore [MWh]"]
        .astype(str)
        .str.replace(",", ".", regex=False)
        .astype(float)
    )
    # Stationary target: first difference
    df["wind_onshore_diff"] = df["wind_onshore"].diff()
    # Lag features from the stationary series
    for lag in range(1, lags + 1):
        df[f"lag_{lag}"] = df["wind_onshore"].shift(lag)
    # Only stationary time features
    df["month"] = df["datetime"].dt.month
    df["dayofweek"] = df["datetime"].dt.dayofweek
    df["hour"] = df["datetime"].dt.hour
    df = df.dropna().reset_index(drop=True)
    feature_cols = ["month", "dayofweek", "hour"] + [
        f"lag_{lag}" for lag in range(1, lags + 1)
    ]
    X = df[feature_cols]
    y = df["wind_onshore_diff"]
    n = len(df)
    n_test = int(n * test_size)

    X_train, X_test = X.iloc[:-n_test], X.iloc[-n_test:]
    y_train, y_test = y.iloc[:-n_test], y.iloc[-n_test:]

    return X_train, X_test, y_train, y_test


if __name__ == "__main__":
    # Try both possible resource paths
    possible_paths = [
        os.path.join("resources", "wind_onshore.csv"),
        os.path.join("..", "..", "resources", "wind_onshore.csv"),
        os.path.join("..", "resources", "wind_onshore.csv"),
        os.path.join(
            os.path.dirname(__file__), "..", "..", "resources", "wind_onshore.csv"
        ),
    ]
    csv_path = None
    for path in possible_paths:
        if os.path.exists(path):
            csv_path = path
            break
    if csv_path is None:
        raise FileNotFoundError(
            "Could not find wind_onshore.csv in any expected location."
        )

    for lags in [0, 1, 2, 5, 10, 20, 50]:
        print(f"\n--- Benchmarking with {lags} lag features ---")
        X_train, X_test, y_train, y_test = load_wind_onshore_stationary(
            csv_path,
            test_size=0.2,
            lags=lags,
        )

        print(f"X_train.cols: {X_train.columns.tolist()}")

        # --- PerpetualBooster ---
        pb = PerpetualBooster(objective="SquaredLoss", budget=1.0)
        start = process_time()
        tick = time()
        pb.fit(X_train, y_train)
        pb_cpu = process_time() - start
        pb_wall = time() - tick
        y_pred_pb = pb.predict(X_test)
        mse_pb = mean_squared_error(y_test, y_pred_pb)
        print(f"PerpetualBooster.n_estimators: {pb.number_of_trees}")
        print(
            f"PerpetualBooster: mse={int(mse_pb)}, cpu={pb_cpu:.2f}s, wall={pb_wall:.2f}s"
        )

        # --- LightGBM ---
        lgbm = LGBMRegressor(n_estimators=100, random_state=42, verbose=-1)
        start = process_time()
        tick = time()
        lgbm.fit(X_train, y_train)
        lgbm_cpu = process_time() - start
        lgbm_wall = time() - tick
        y_pred_lgbm = lgbm.predict(X_test)
        mse_lgbm = mean_squared_error(y_test, y_pred_lgbm)
        print(
            f"LightGBM: mse={int(mse_lgbm)}, cpu={lgbm_cpu:.2f}s, wall={lgbm_wall:.2f}s"
        )



"""Policy learning via Inverse Propensity Weighting (IPW / AIPW).

Implements the Athey & Wager (2021) policy-learning framework using
gradient boosting.  The learned policy assigns treatment when the
boosted score is positive.

Two modes are supported:

* **IPW** — Standard Inverse Propensity Weighting.
* **AIPW** (Augmented / Doubly Robust) — Reduces variance by
  incorporating a baseline outcome model.
"""

from typing import Optional

import numpy as np
from typing_extensions import Self

from perpetual.booster import PerpetualBooster
from perpetual.perpetual import PolicyObjective

# ---------------------------------------------------------------------------
# Custom objective helpers
# ---------------------------------------------------------------------------


# ---------------------------------------------------------------------------
# PolicyLearner
# ---------------------------------------------------------------------------


class PolicyLearner:
    r"""Policy learner via Inverse Propensity Weighting.

    Learns a treatment-assignment policy :math:`\\pi(X)` that maximizes
    expected reward using the Athey & Wager (2021) policy-learning
    framework.

    The learned policy assigns :math:`W = 1` when the boosted
    score :math:`F(X) > 0`.

    Parameters
    ----------
    budget : float, default=0.5
        Fitting budget forwarded to ``PerpetualBooster``.
    mode : str, default="ipw"
        ``"ipw"`` for standard Inverse Propensity Weighting or
        ``"aipw"`` for Augmented (Doubly Robust) IPW.
    propensity_budget : float, optional
        Separate budget for the propensity model.  If ``None``, defaults to
        ``budget``.  Only used when ``propensity`` is not supplied to
        :meth:`fit`.
    **kwargs
        Additional keyword arguments forwarded to ``PerpetualBooster``.

    Attributes
    ----------
    feature_importances_ : ndarray of shape (n_features,)
        Feature importances from the policy model.

    Examples
    --------
    >>> from perpetual.policy import PolicyLearner
    >>> import numpy as np
    >>> n = 500
    >>> X = np.random.randn(n, 5)
    >>> w = np.random.binomial(1, 0.5, n)
    >>> y = X[:, 0] * w + np.random.randn(n) * 0.5
    >>> pl = PolicyLearner(budget=0.3)
    >>> pl.fit(X, w, y)  # doctest: +SKIP
    >>> policy = pl.predict(X)  # doctest: +SKIP

    References
    ----------
    Athey, S., & Wager, S. (2021). *Policy learning with observational data*.
    Econometrica, 89(1), 133-161.
    """

    def __init__(
        self,
        budget: float = 0.5,
        mode: str = "ipw",
        propensity_budget: Optional[float] = None,
        **kwargs,
    ):
        if mode not in ("ipw", "aipw"):
            raise ValueError(f"mode must be 'ipw' or 'aipw', got {mode!r}.")
        self.budget = budget
        self.mode = mode
        self.propensity_budget = propensity_budget or budget
        self._kwargs = kwargs
        self.feature_importances_: Optional[np.ndarray] = None

    def fit(
        self,
        X,
        w,
        y,
        propensity: Optional[np.ndarray] = None,
        mu_hat_1: Optional[np.ndarray] = None,
        mu_hat_0: Optional[np.ndarray] = None,
    ) -> Self:
        r"""Fit the policy learner.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Covariate matrix.
        w : array-like of shape (n_samples,)
            Observed binary treatment assignment (0 or 1).
        y : array-like of shape (n_samples,)
            Observed outcome.
        propensity : array-like of shape (n_samples,), optional
            Estimated :math:`P(W=1|X)`.  If ``None``, a propensity model is
            fitted internally.
        mu_hat_1 : array-like of shape (n_samples,), optional
            Predicted outcome under treatment :math:`\hat{\mu}_1(X)`.
            Required when ``mode="aipw"``.
        mu_hat_0 : array-like of shape (n_samples,), optional
            Predicted outcome under control :math:`\hat{\mu}_0(X)`.
            Required when ``mode="aipw"``.

        Returns
        -------
        self
            Fitted estimator.
        """
        X_arr = np.asarray(X, dtype=float)
        w_arr = np.asarray(w, dtype=float)
        y_arr = np.asarray(y, dtype=float)

        # Propensity scores
        if propensity is not None:
            p_hat = np.asarray(propensity, dtype=float)
        else:
            prop_model = PerpetualBooster(
                budget=self.propensity_budget, objective="LogLoss", **self._kwargs
            )
            prop_model.fit(X_arr, w_arr)
            log_odds = prop_model.predict(X_arr)
            p_hat = 1.0 / (1.0 + np.exp(-log_odds))

        # Baseline outcome models for AIPW
        m1 = None
        m0 = None
        if self.mode == "aipw":
            if mu_hat_1 is not None and mu_hat_0 is not None:
                m1 = np.asarray(mu_hat_1, dtype=float)
                m0 = np.asarray(mu_hat_0, dtype=float)
            else:
                # Fit separate outcome models for treatment and control
                m1_model = PerpetualBooster(
                    budget=self.budget, objective="SquaredLoss", **self._kwargs
                )
                m0_model = PerpetualBooster(
                    budget=self.budget, objective="SquaredLoss", **self._kwargs
                )
                mask1 = w_arr == 1
                mask0 = w_arr == 0
                if mask1.any():
                    m1_model.fit(X_arr[mask1], y_arr[mask1])
                if mask0.any():
                    m0_model.fit(X_arr[mask0], y_arr[mask0])

                m1 = m1_model.predict(X_arr)
                m0 = m0_model.predict(X_arr)

        # Policy model via Rust PolicyObjective
        objective = PolicyObjective(
            treatment=w_arr.astype("uint8"),
            propensity=p_hat,
            mode=self.mode,
            mu_hat_1=m1,
            mu_hat_0=m0,
        )

        self._policy_model = PerpetualBooster(
            budget=self.budget, objective=objective, **self._kwargs
        )
        # Pass observed outcome y_arr. The definition of Gamma is handled internally
        # by the RustPolicyObjective, which implements pseudo_outcome using y.
        self._policy_model.fit(X_arr, y_arr)
        self.feature_importances_ = self._policy_model.feature_importances_

        return self

    def predict(self, X) -> np.ndarray:
        """Predict the optimal treatment assignment.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Covariate matrix.

        Returns
        -------
        ndarray of shape (n_samples,)
            Binary treatment policy (1 = treat, 0 = do not treat).
        """
        scores = self.decision_function(X)
        return (scores > 0).astype(int)

    def decision_function(self, X) -> np.ndarray:
        """Return raw policy scores.

        Positive values indicate treatment is beneficial.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Covariate matrix.

        Returns
        -------
        ndarray of shape (n_samples,)
            Raw boosted policy scores.
        """
        return self._policy_model.predict(X)

    def predict_proba(self, X) -> np.ndarray:
        """Predict probability that treatment is beneficial.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Covariate matrix.

        Returns
        -------
        ndarray of shape (n_samples,)
            Probability of treatment being beneficial (sigmoid of score).
        """
        scores = self.decision_function(X)
        probabilities = np.empty_like(scores, dtype=np.float64)
        positive = scores >= 0.0
        probabilities[positive] = 1.0 / (1.0 + np.exp(-scores[positive]))
        exp_scores = np.exp(scores[~positive])
        probabilities[~positive] = exp_scores / (1.0 + exp_scores)
        return probabilities
